# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# Changes Made from original:
#   import paths
# ******************************************************************************
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utility functions specifically for NMT."""
from __future__ import print_function

import codecs
import time
import numpy as np
import tensorflow as tf

from nlp_architect.models.gnmt.utils import evaluation_utils
from nlp_architect.models.gnmt.utils import misc_utils as utils

__all__ = ["decode_and_evaluate", "get_translation"]


def decode_and_evaluate(name,
                        model,
                        sess,
                        trans_file,
                        ref_file,
                        metrics,
                        subword_option,
                        beam_width,
                        tgt_eos,
                        num_translations_per_input=1,
                        decode=True,
                        infer_mode="greedy"):
    """Decode a test set and compute a score according to the evaluation task."""
    # Decode
    if decode:
        utils.print_out("  decoding to output %s" % trans_file)

        start_time = time.time()
        num_sentences = 0
        with codecs.getwriter("utf-8")(
                tf.gfile.GFile(trans_file, mode="wb")) as trans_f:
            trans_f.write("")  # Write empty string to ensure file is created.

            if infer_mode == "greedy":
                num_translations_per_input = 1
            elif infer_mode == "beam_search":
                num_translations_per_input = min(num_translations_per_input, beam_width)

            while True:
                try:
                    nmt_outputs, _ = model.decode(sess)
                    if infer_mode != "beam_search":
                        nmt_outputs = np.expand_dims(nmt_outputs, 0)

                    batch_size = nmt_outputs.shape[1]
                    num_sentences += batch_size

                    for sent_id in range(batch_size):
                        for beam_id in range(num_translations_per_input):
                            translation = get_translation(
                                nmt_outputs[beam_id],
                                sent_id,
                                tgt_eos=tgt_eos,
                                subword_option=subword_option)
                            trans_f.write((translation + b"\n").decode("utf-8"))
                except tf.errors.OutOfRangeError:
                    utils.print_time(
                        "  done, num sentences %d, num translations per input %d" %
                        (num_sentences, num_translations_per_input), start_time)
                    break

    # Evaluation
    evaluation_scores = {}
    if ref_file and tf.gfile.Exists(trans_file):
        for metric in metrics:
            score = evaluation_utils.evaluate(
                ref_file,
                trans_file,
                metric,
                subword_option=subword_option)
            evaluation_scores[metric] = score
            utils.print_out("  %s %s: %.1f" % (metric, name, score))

    return evaluation_scores


def get_translation(nmt_outputs, sent_id, tgt_eos, subword_option):
    """Given batch decoding outputs, select a sentence and turn to text."""
    if tgt_eos:
        tgt_eos = tgt_eos.encode("utf-8")
    # Select a sentence
    output = nmt_outputs[sent_id, :].tolist()

    # If there is an eos symbol in outputs, cut them at that point.
    if tgt_eos and tgt_eos in output:
        output = output[:output.index(tgt_eos)]

    if subword_option == "bpe":  # BPE
        translation = utils.format_bpe_text(output)
    elif subword_option == "spm":  # SPM
        translation = utils.format_spm_text(output)
    else:
        translation = utils.format_text(output)

    return translation



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf


def simple_lstm(max_features, dense_out, input_length, embed_dim=256, lstm_out=140,
                dropout=0.5):
    """
    Simple Bi-direction LSTM Model in Keras

    Single layer bi-directional lstm with recurrent dropout and a fully connected layer

    Args:
        max_features (int): vocabulary size
        dense_out (int): size out the output dense layer, this is the number of classes
        input_length (int): length of the input text
        embed_dim (int): internal embedding size used in the lstm
        lstm_out (int): size of the bi-directional output layer
        dropout (float): value for recurrent dropout, between 0 and 1

    Returns:
        model (model): LSTM model
    """
    model = tf.keras.models.Sequential()
    model.add(tf.keras.layers.Embedding(max_features, embed_dim, input_length=input_length))
    model.add(tf.keras.layers.Bidirectional(
        tf.keras.layers.LSTM(lstm_out, recurrent_dropout=dropout, activation='tanh')))
    model.add(tf.keras.layers.Dense(dense_out, activation='softmax'))
    model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])

    return model


def one_hot_cnn(dense_out, max_len=300, frame='small'):
    """
    Temporal CNN Model

    As defined in "Text Understanding from Scratch" by Zhang, LeCun 2015
    https://arxiv.org/pdf/1502.01710v4.pdf
    This model is a series of 1D CNNs, with a maxpooling and fully connected layers.
    The frame sizes may either be large or small.


    Args:
        dense_out (int): size out the output dense layer, this is the number of classes
        max_len (int): length of the input text
        frame (str): frame size, either large or small

    Returns:
        model (model): temporal CNN model
    """

    if frame == 'large':
        cnn_size = 1024
        fully_connected = [2048, 2048, dense_out]
    else:
        cnn_size = 256
        fully_connected = [1024, 1024, dense_out]

    model = tf.keras.models.Sequential()

    model.add(tf.keras.layers.Conv1D(cnn_size, 7, padding='same', input_shape=(68, max_len)))
    model.add(tf.keras.layers.MaxPooling1D(pool_size=3))

    print(model.output_shape)

    # Input = 22 x 256
    model.add(tf.keras.layers.Conv1D(cnn_size, 7, padding='same'))
    model.add(tf.keras.layers.MaxPooling1D(pool_size=3))

    print(model.output_shape)
    # Input = 7 x 256
    model.add(tf.keras.layers.Conv1D(cnn_size, 3, padding='same'))

    # Input = 7 x 256
    model.add(tf.keras.layers.Conv1D(cnn_size, 3, padding='same'))

    model.add(tf.keras.layers.Conv1D(cnn_size, 3, padding='same'))

    # Input = 7 x 256
    model.add(tf.keras.layers.Conv1D(cnn_size, 3, padding='same'))
    model.add(tf.keras.layers.MaxPooling1D(pool_size=3))

    model.add(tf.keras.layers.Flatten())

    # Fully Connected Layers

    # Input is 512 Output is 1024/2048
    model.add(tf.keras.layers.Dense(fully_connected[0]))
    model.add(tf.keras.layers.Dropout(0.75))
    model.add(tf.keras.layers.Activation('relu'))

    # Input is 1024/2048 Output is 1024/2048
    model.add(tf.keras.layers.Dense(fully_connected[1]))
    model.add(tf.keras.layers.Dropout(0.75))
    model.add(tf.keras.layers.Activation('relu'))

    # Input is 1024/2048 Output is dense_out size (number of classes)
    model.add(tf.keras.layers.Dense(fully_connected[2]))
    model.add(tf.keras.layers.Activation('softmax'))

    # Stochastic gradient parameters as set by paper
    sgd = tf.keras.optimizers.SGD(lr=0.01, decay=1e-5, momentum=0.9, nesterov=True)
    model.compile(loss='categorical_crossentropy', optimizer=sgd, metrics=['accuracy'])

    return model



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import sys
from typing import List

from nlp_architect.utils.io import load_json_file
from nlp_architect.utils.string_utils import StringUtils


class MentionDataLight(object):
    def __init__(self, tokens_str: str, mention_context: str = None, mention_head: str = None,
                 mention_head_lemma: str = None, mention_pos: str = None,
                 mention_ner: str = None):
        """
        Object represent a mention with only text values
        Args:
            tokens_str: str the tokens combine text (join with space)
            mention_head: str
            mention_head_lemma: str
        """
        self.tokens_str = tokens_str
        self.mention_context = mention_context
        if not mention_head and not mention_head_lemma:
            self.mention_head, self.mention_head_lemma, self.mention_head_pos, \
                self.mention_ner = StringUtils.find_head_lemma_pos_ner(str(tokens_str))
        else:
            self.mention_head = mention_head
            self.mention_head_lemma = mention_head_lemma
            self.mention_head_pos = mention_pos
            self.mention_ner = mention_ner


class MentionData(MentionDataLight):
    def __init__(self, topic_id: str, doc_id: str, sent_id: int, tokens_numbers: List[int],
                 tokens_str: str, mention_context: List[str], mention_head: str,
                 mention_head_lemma: str, coref_chain: str, mention_type: str = 'NA',
                 is_continuous: bool = True, is_singleton: bool = False, score: float = float(-1),
                 predicted_coref_chain: str = None, mention_pos: str = None,
                 mention_ner: str = None, mention_index: int = -1) -> None:
        """
        Object represent a mention

        Args:
            topic_id: str topic ID
            doc_id: str document ID
            sent_id: int sentence number
            tokens_numbers: List[int] - tokens numbers
            mention_context: List[str] - list of tokens strings
            coref_chain: str
            mention_type: str one of (HUM/NON/TIM/LOC/ACT/NEG)
            is_continuous: bool
            is_singleton: bool
            score: float
            predicted_coref_chain: str (should be field while evaluated)
            mention_pos: str
            mention_ner: str
            mention_index: in case order is of value (default = -1)
        """
        super(MentionData, self).__init__(tokens_str, mention_context, mention_head,
                                          mention_head_lemma, mention_pos,
                                          mention_ner)
        self.topic_id = topic_id
        self.doc_id = doc_id
        self.sent_id = sent_id
        self.tokens_number = tokens_numbers
        self.mention_type = mention_type
        self.coref_chain = coref_chain
        self.is_continuous = is_continuous
        self.is_singleton = is_singleton
        self.score = score
        self.predicted_coref_chain = predicted_coref_chain
        self.mention_id = self.gen_mention_id()
        self.mention_index = mention_index

    @staticmethod
    def read_json_mention_data_line(mention_line: str):
        """
        Args:
            mention_line: a Json representation of a single mention

        Returns:
            MentionData object
        """
        # pylint: disable=too-many-branches

        try:
            topic_id = None
            coref_chain = None
            doc_id = None
            sent_id = None
            tokens_numbers = None
            score = -1
            mention_type = None
            predicted_coref_chain = None
            mention_context = None
            is_continue = False
            is_singleton = False
            mention_pos = None
            mention_ner = None
            mention_index = -1

            mention_text = mention_line['tokens_str']

            if 'topic_id' in mention_line:
                topic_id = mention_line['topic_id']

            if 'coref_chain' in mention_line:
                coref_chain = mention_line['coref_chain']

            if 'doc_id' in mention_line:
                doc_id = mention_line['doc_id']
                if '.xml' not in doc_id:
                    doc_id = doc_id + '.xml'

            if 'sent_id' in mention_line:
                sent_id = mention_line['sent_id']

            if 'tokens_number' in mention_line:
                tokens_numbers = mention_line['tokens_number']

            if 'mention_context' in mention_line:
                mention_context = mention_line['mention_context']

            if 'mention_head' in mention_line and 'mention_head_lemma' in mention_line:
                mention_head = mention_line['mention_head']
                mention_head_lemma = mention_line['mention_head_lemma']
                if 'mention_head_pos' in mention_line:
                    mention_pos = mention_line['mention_head_pos']
                if 'mention_ner' in mention_line:
                    mention_ner = mention_line['mention_ner']
            else:
                mention_head, mention_head_lemma, mention_pos, \
                    mention_ner = StringUtils.find_head_lemma_pos_ner(str(mention_text))

            if 'mention_type' in mention_line:
                mention_type = mention_line['mention_type']
            if 'score' in mention_line:
                score = mention_line['score']

            if 'is_continuous' in mention_line:
                is_continue = mention_line['is_continuous']

            if 'is_singleton' in mention_line:
                is_singleton = mention_line['is_singleton']

            if 'predicted_coref_chain' in mention_line:
                predicted_coref_chain = mention_line['predicted_coref_chain']

            if 'mention_index' in mention_line:
                mention_index = mention_line['mention_index']

            mention_data = MentionData(topic_id, doc_id, sent_id, tokens_numbers, mention_text,
                                       mention_context,
                                       mention_head, mention_head_lemma,
                                       coref_chain, mention_type, is_continue, is_singleton, score,
                                       predicted_coref_chain, mention_pos, mention_ner,
                                       mention_index)
        except Exception:
            print('Unexpected error:', sys.exc_info()[0])
            raise Exception('failed reading json line-' + str(mention_line))

        return mention_data

    @staticmethod
    def read_mentions_json_to_mentions_data_list(mentions_json_file: str):
        """

        Args:
            mentions_json_file: the path of the mentions json file to read

        Returns:
            List[MentionData]
        """
        all_mentions_only = load_json_file(mentions_json_file)

        mentions = []
        for mention_line in all_mentions_only:
            mention_data = MentionData.read_json_mention_data_line(mention_line)
            mentions.append(mention_data)

        return mentions

    def get_tokens(self):
        return self.tokens_number

    def gen_mention_id(self) -> str:
        if self.doc_id and self.sent_id is not None and self.tokens_number:
            tokens_ids = [str(self.doc_id), str(self.sent_id)]
            tokens_ids.extend([str(token_id) for token_id in self.tokens_number])
            return '_'.join(tokens_ids)

        return '_'.join(self.tokens_str.split())

    def get_mention_id(self) -> str:
        if not self.mention_id:
            self.mention_id = self.gen_mention_id()
        return self.mention_id

    @staticmethod
    def static_gen_token_unique_id(doc_id: int, sent_id: int, token_id: int) -> str:
        return '_'.join([str(doc_id), str(sent_id), str(token_id)])



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import json
from enum import Enum
from json import JSONEncoder


class LexiconElement(object):
    def __init__(self, term: list, score: str or float = None, polarity: str = None,
                 is_acquired: str = None, position: str = None):
        self.term = term
        self.polarity = polarity
        try:
            self.score = float(score)
        except TypeError:
            self.score = 0
        self.position = position
        if is_acquired == "N":
            self.is_acquired = False
        elif is_acquired == "Y":
            self.is_acquired = True
        else:
            self.is_acquired = None

    def __lt__(self, other):
        return self.term[0] < other.term[0]

    def __le__(self, other):
        return self.term[0] <= other.term[0]

    def __eq__(self, other):
        return self.term[0] == other.term[0]

    def __ne__(self, other):
        return self.term[0] != other.term[0]

    def __gt__(self, other):
        return self.term[0] > other.term[0]

    def __ge__(self, other):
        return self.term[0] >= other.term[0]


class TermType(Enum):
    OPINION = 'OP'
    ASPECT = 'AS'
    NEGATION = 'NEG'
    INTENSIFIER = 'INT'


class Polarity(Enum):
    POS = 'POS'
    NEG = 'NEG'
    UNK = 'UNK'


class Term(object):
    def __init__(self, text: str, kind: TermType, polarity: Polarity, score: float, start: int,
                 length: int):
        self._text = text
        self._type = kind
        self._polarity = polarity
        self._score = score
        self._start = start
        self._len = length

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    @property
    def text(self):
        return self._text

    @property
    def type(self):
        return self._type

    @property
    def polarity(self):
        return self._polarity

    @property
    def score(self):
        return self._score

    @property
    def start(self):
        return self._start

    @property
    def len(self):
        return self._len

    @text.setter
    def text(self, val):
        self._text = val

    @score.setter
    def score(self, val):
        self._score = val

    @polarity.setter
    def polarity(self, val):
        self._polarity = val

    def __str__(self):
        return "text: " + self._text + " type: " + str(self._type) + " pol: " + \
               str(self._polarity) + " score: " + str(self._score) + " start: " + \
               str(self._start) + " len: " + \
               str(self._len)


class SentimentDoc(object):
    def __init__(self, doc_text: str = None, sentences: list = None):
        if sentences is None:
            sentences = []
        self._doc_text = doc_text
        self._sentences = sentences

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    @property
    def doc_text(self):
        return self._doc_text

    @doc_text.setter
    def doc_text(self, val):
        self._doc_text = val

    @property
    def sentences(self):
        return self._sentences

    @sentences.setter
    def sentences(self, val):
        self._sentences = val

    @staticmethod
    def decoder(obj):
        """
        :param obj: object to be decoded
        :return: decoded Sentence object
        """
        # SentimentDoc
        if '_doc_text' in obj and '_sentences' in obj:
            return SentimentDoc(obj['_doc_text'], obj['_sentences'])

        # SentimentSentence
        if all((attr in obj for attr in ('_start', '_end', '_events'))):
            return SentimentSentence(obj['_start'], obj['_end'], obj['_events'])

        # Term
        if all(attr in obj for attr in
               ('_text', '_type', '_score', '_polarity', '_start', '_len')):
            return Term(obj['_text'], TermType[obj['_type']],
                        Polarity[obj['_polarity']], obj['_score'], obj['_start'],
                        obj['_len'])
        return obj

    def __repr__(self):
        return self.pretty_json()

    def __str__(self):
        return self.__repr__()

    def __iter__(self):
        return self.sentences.__iter__()

    def __len__(self):
        return len(self.sentences)

    def json(self):
        """
        Return json representations of the object

        Returns:
            :obj:`json`: json representations of the object
        """
        return json.dumps(self, cls=SentimentDocEncoder)

    def pretty_json(self):
        """
        Return pretty json representations of the object

        Returns:
            :obj:`json`: pretty json representations of the object
        """
        return json.dumps(self, cls=SentimentDocEncoder, indent=4)


class SentimentSentence(object):
    def __init__(self, start: int, end: int, events: list):
        self._start = start
        self._end = end
        self._events = events

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, val):
        self._start = val

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, val):
        self._end = val

    @property
    def events(self):
        return self._events

    @events.setter
    def events(self, val):
        self._events = val


class SentimentDocEncoder(JSONEncoder):
    def default(self, o):  # pylint: disable=E0202
        try:
            if isinstance(o, Enum):
                return getattr(o, '_name_')
            if hasattr(o, '__dict__'):
                return vars(o)
            if hasattr(o, '__slots__'):
                ret = {slot: getattr(o, slot) for slot in o.__slots__}
                for cls in type(o).mro():
                    spr = super(cls, o)
                    if not hasattr(spr, '__slots__'):
                        break
                    for slot in spr.__slots__:
                        ret[slot] = getattr(o, slot)
                return ret
        except Exception as e:
            print(e)
        return None



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import argparse
import json
from os import PathLike
from os.path import isdir
from typing import Optional

import numpy as np
import pandas as pd
from tqdm import tqdm

from nlp_architect.common.core_nlp_doc import CoreNLPDoc
from nlp_architect.models.absa.inference.data_types import TermType, \
    SentimentDocEncoder, SentimentDoc
from nlp_architect.models.absa.inference.inference \
    import SentimentInference
from nlp_architect.models.absa.utils import load_opinion_lex
from nlp_architect.solutions.absa_solution import SENTIMENT_OUT
from nlp_architect.solutions.absa_solution.ui import serve_ui, _ui_format
from nlp_architect.solutions.absa_solution.utils import Anonymiser
from nlp_architect.utils.io import walk_directory, validate_existing_filepath, \
    validate_existing_directory


class SentimentSolution(object):
    """Main class for executing Sentiment Solution pipeline.

    Args:
        anonymiser (Anonymiser, optional): Method to anonymise events' text.
        max_events (int, optional): Maximum number of events to show for each aspect-polarity pair.
    """

    def __init__(self, anonymiser: Anonymiser = None, max_events: int = 400):
        self.anonymiser = anonymiser
        self.max_events = max_events
        SENTIMENT_OUT.mkdir(parents=True, exist_ok=True)

    def run(self, aspect_lex: PathLike = None, opinion_lex: PathLike = None,
            data: PathLike = None, parsed_data: PathLike = None,
            inference_results: PathLike = None, ui=True) -> Optional[pd.DataFrame]:

        opinions = load_opinion_lex(opinion_lex)
        if not opinions:
            raise ValueError('Empty opinion lexicon!')
        aspects = pd.read_csv(aspect_lex, header=None, encoding='utf-8')[0]
        if aspects.empty:
            raise ValueError('Empty aspect lexicon!')
        if inference_results:
            with open(inference_results) as f:
                results = json.loads(f.read(), object_hook=SentimentDoc.decoder)
        elif data or parsed_data:
            inference = SentimentInference(aspect_lex, opinions, parse=False)
            parse = None
            if not parsed_data:  # source data is raw text, need to parse
                from nlp_architect.pipelines.spacy_bist import SpacyBISTParser
                parse = SpacyBISTParser().parse

            results = {}
            print('Running inference on data files... (Iterating data files)')
            data_source = parsed_data if parsed_data else data
            for file, doc in self._iterate_docs(data_source):
                parsed_doc = parse(doc) if parse \
                    else json.loads(doc, object_hook=CoreNLPDoc.decoder)
                sentiment_doc = inference.run(parsed_doc=parsed_doc)
                if sentiment_doc:
                    results[file] = sentiment_doc
            with open(SENTIMENT_OUT / 'inference_results.json', 'w') as f:
                json.dump(results, f, cls=SentimentDocEncoder, indent=4, sort_keys=True)
        else:
            print('No input given. Please supply one of: '
                  'data directory, parsed data directory, or inference results.')
            return None

        print("\nComputing statistics...")
        stats = self._compute_stats(results, aspects, opinions)
        print("Done.")
        if ui:
            serve_ui(stats, aspects)
        return stats

    @staticmethod
    def _iterate_docs(data: PathLike) -> tuple:
        if isdir(data):
            for file, doc_text in tqdm(list(walk_directory(data))):
                yield file, doc_text
        else:
            with open(data, encoding='utf-8') as f:
                for i, doc_text in tqdm(enumerate(f), total=_line_count(data)):
                    yield str(i + 1), doc_text

    def _compute_stats(self, results: dict, aspects: list, opinion_lex: dict) -> pd.DataFrame:
        """Aggregates counts for each aspect-polarity pairs, with separate counts for in-domain
         only events.
        """
        index = pd.MultiIndex.from_product([aspects, ['POS', 'NEG'], [False, True]],
                                           names=['Aspect', 'Polarity', 'inDomain'])
        stats = pd.DataFrame(columns=['Quantity', 'Score'], index=index)
        stats[['Quantity', 'Score']] = stats[['Quantity', 'Score']].fillna(0)
        stats = stats.sort_index()
        scores = stats.copy()

        for doc in tqdm(results.values()):
            for sent in doc.sentences:
                for event in sent.events:
                    aspect = [t for t in event if t.type == TermType.ASPECT][0]
                    opinion = [t for t in event if t.type == TermType.OPINION][0]
                    score = aspect.score
                    key = aspect.text, aspect.polarity.name
                    count = self._add_event(stats, key, False, score)
                    in_domain = opinion_lex[opinion.text.lower()].is_acquired
                    count_dom = self._add_event(stats, key, True, score) if in_domain else -1

                    if count <= self.max_events:
                        sent_ui = _ui_format(sent, doc)
                        self._add_sentence(sent_ui, stats, scores, key, False, count, score)
                        if in_domain:
                            self._add_sentence(sent_ui, stats, scores, key, True, count_dom, score)
        for key in index:  # sort sentences according to their scores
            stats.loc[key, 2:] = stats.loc[key][2:][np.argsort(scores.loc[key][2:])].tolist()
        return stats

    def _add_sentence(self, sent_ui: str, stats: pd.DataFrame, scores: pd.DataFrame, key: tuple,
                      in_domain: bool, count: int, score: int) -> int:
        """Utility function for adding event sentence to output."""
        sent_ui = self.anonymiser.run(sent_ui) if self.anonymiser else sent_ui
        sent_key = key + (in_domain,), 'Sent_' + str(count)
        stats.at[sent_key] = sent_ui
        scores.at[sent_key] = -abs(score)
        return count

    @staticmethod
    def _add_event(df: pd.DataFrame, key: tuple, in_domain: bool, score: int) -> int:
        """Utility function for incrementing event counts."""
        key = key + (in_domain,)
        count = int(df.loc[key, 'Quantity']) + 1
        df.loc[key, 'Quantity'] = count
        df.loc[key, 'Score'] += score
        return count


def _line_count(file):
    """Utility function for getting number of lines in a text file."""
    with open(file) as f:
        return len(list(f))


def main() -> None:
    parser = argparse.ArgumentParser(description='Aspect-Based Sentiment Analysis')
    parser.add_argument('--data', type=validate_existing_directory,
                        help='Path to data')
    parser.add_argument('--aspects', type=validate_existing_filepath,
                        help='Path to aspect lexicon', required=True)
    parser.add_argument('--opinions', type=validate_existing_filepath,
                        help='Path to opinion lexicon', required=True)
    parser.add_argument('--parsed', type=validate_existing_directory,
                        help='Path to parsed data')
    parser.add_argument('--res', type=validate_existing_filepath,
                        help='Path to inference results')
    args = parser.parse_args()

    solution = SentimentSolution()
    solution.run(data=args.data,
                 parsed_data=args.parsed,
                 inference_results=args.res,
                 aspect_lex=args.aspects,
                 opinion_lex=args.opinions)


if __name__ == '__main__':
    main()



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import json
import logging
import sys

from gensim.models import FastText, Word2Vec, KeyedVectors
from gensim.models.word2vec import LineSentence
from gensim import utils
import nltk
from nltk.corpus import conll2000
from six import iteritems

logger = logging.getLogger(__name__)


# pylint: disable-msg=too-many-instance-attributes
class NP2vec:
    """
    Initialize the np2vec model, train it, save it and load it.
    """

    def is_marked(self, s):
        """
        Check if a string is marked.

        Args:
            s (str): string to check
        """
        return len(s) > 0 and s[-1] == self.mark_char

    # pylint: disable-msg=too-many-arguments
    # pylint: disable-msg=too-many-locals
    # pylint: disable-msg=too-many-branches
    def __init__(  # noqa: C901
            self,
            corpus,
            corpus_format='txt',
            mark_char='_',
            word_embedding_type='word2vec',
            sg=0,
            size=100,
            window=10,
            alpha=0.025,
            min_alpha=0.0001,
            min_count=5,
            sample=1e-5,
            workers=20,
            hs=0,
            negative=25,
            cbow_mean=1,
            iterations=15,
            min_n=3,
            max_n=6,
            word_ngrams=1,
            prune_non_np=True):
        """
        Initialize np2vec model and train it.

        Args:
          corpus (str): path to the corpus.
          corpus_format (str {json,txt,conll2000}): format of the input marked corpus; txt and json
          formats are supported. For json format, the file should contain an iterable of
          sentences. Each sentence is a list of terms (unicode strings) that will be used for
          training.
          mark_char (char): special character that marks NP's suffix.
          word_embedding_type (str {word2vec,fasttext}): word embedding model type; word2vec and
          fasttext are supported.
          np2vec_model_file (str): path to the file where the trained np2vec model has to be
          stored.
          binary (bool): boolean indicating whether the model is stored in binary format; if
          word_embedding_type is fasttext and word_ngrams is 1, binary should be set to True.
          sg (int {0,1}): model training hyperparameter, skip-gram. Defines the training
          algorithm. If 1, CBOW is used,otherwise, skip-gram is employed.
          size (int): model training hyperparameter, size of the feature vectors.
          window (int): model training hyperparameter, maximum distance between the current and
          predicted word within a sentence.
          alpha (float): model training hyperparameter. The initial learning rate.
          min_alpha (float): model training hyperparameter. Learning rate will linearly drop to
          `min_alpha` as training progresses.
          min_count (int): model training hyperparameter, ignore all words with total frequency
          lower than this.
          sample (float): model training hyperparameter, threshold for configuring which
          higher-frequency words are randomly downsampled, useful range is (0, 1e-5)
          workers (int): model training hyperparameter, number of worker threads.
          hs (int {0,1}): model training hyperparameter, hierarchical softmax. If set to 1,
          hierarchical softmax will be used for model training. If set to 0, and `negative` is non-
                        zero, negative sampling will be used.
          negative (int): model training hyperparameter, negative sampling. If > 0, negative
          sampling will be used, the int for negative specifies how many "noise words" should be
          drawn (usually between 5-20). If set to 0, no negative sampling is used.
          cbow_mean (int {0,1}): model training hyperparameter. If 0, use the sum of the context
          word vectors. If 1, use the mean, only applies when cbow is used.
          iterations (int): model training hyperparameter, number of iterations.
          min_n (int): fasttext training hyperparameter. Min length of char ngrams to be used
          for training word representations.
          max_n (int): fasttext training hyperparameter. Max length of char ngrams to be used for
          training word representations. Set `max_n` to be lesser than `min_n` to avoid char
          ngrams being used.
          word_ngrams (int {0,1}): fasttext training hyperparameter. If 1, uses enrich word
          vectors with subword (ngrams) information. If 0, this is equivalent to word2vec training.
          prune_non_np (bool): indicates whether to prune non-NP's after training process.

        """

        self.mark_char = mark_char
        self.word_embedding_type = word_embedding_type
        self.sg = sg
        self.size = size
        self.window = window
        self.alpha = alpha
        self.min_alpha = min_alpha
        self.min_count = min_count
        self.sample = sample
        self.workers = workers
        self.hs = hs
        self.negative = negative
        self.cbow_mean = cbow_mean
        self.iter = iterations
        self.min_n = min_n
        self.max_n = max_n
        self.word_ngrams = word_ngrams
        self.prune_non_np = prune_non_np

        if corpus_format == 'txt':
            self._sentences = LineSentence(corpus)
        elif corpus_format == 'json':
            with open(corpus) as json_data:
                self._sentences = json.load(json_data)
        # pylint: disable-msg=too-many-nested-blocks
        elif corpus_format == 'conll2000':
            try:
                self._sentences = list()
                for chunked_sent in conll2000.chunked_sents(corpus):
                    tokens = list()
                    for chunk in chunked_sent:
                        # pylint: disable-msg=protected-access
                        if hasattr(chunk, '_label') and chunk._label == 'NP':
                            s = ''
                            for w in chunk:
                                s += w[0] + self.mark_char
                            tokens.append(s)
                        else:
                            if isinstance(chunk, nltk.Tree):
                                for w in chunk:
                                    tokens.append(w[0])
                            else:
                                tokens.append(chunk[0])
                        self._sentences.append(tokens)
            # pylint: disable-msg=broad-except
            except Exception:
                print('Conll2000 dataset is missing. See downloading details in the '
                      'README file')
        else:
            logger.error('invalid corpus format: %s', corpus_format)
            sys.exit(0)

        if word_embedding_type == 'fasttext' and word_ngrams == 1:
            # remove the marking character at the end for subword fasttext model training
            self._sentences = [[w[:-1] if self.is_marked(w) else w for w in sentence]
                               for sentence in self._sentences]

        logger.info('training np2vec model')
        self._train()

    def _train(self):
        """
        Train the np2vec model.
        """
        if self.word_embedding_type == 'word2vec':
            self.model = Word2Vec(
                self._sentences,
                sg=self.sg,
                size=self.size,
                window=self.window,
                alpha=self.alpha,
                min_alpha=self.min_alpha,
                min_count=self.min_count,
                sample=self.sample,
                workers=self.workers,
                hs=self.hs,
                negative=self.negative,
                cbow_mean=self.cbow_mean,
                iter=self.iter)

        elif self.word_embedding_type == 'fasttext':
            self.model = FastText(
                self._sentences,
                sg=self.sg,
                size=self.size,
                window=self.window,
                alpha=self.alpha,
                min_alpha=self.min_alpha,
                min_count=self.min_count,
                sample=self.sample,
                workers=self.workers,
                hs=self.hs,
                negative=self.negative,
                cbow_mean=self.cbow_mean,
                iter=self.iter,
                min_n=self.min_n,
                max_n=self.max_n,
                word_ngrams=self.word_ngrams)
        else:
            logger.error('invalid word embedding type: %s', self.word_embedding_type)
            sys.exit(0)

    def save(self, np2vec_model_file='np2vec.model', binary=False, word2vec_format=True):
        """
        Save the np2vec model.

        Args:
            np2vec_model_file (str): the file containing the np2vec model to load
            binary (bool): boolean indicating whether the np2vec model to load is in binary format
            word2vec_format(bool): boolean indicating whether to save the model in original
            word2vec format.
        """
        if self.word_embedding_type == 'fasttext' and self.word_ngrams == 1:
            if not binary:
                logger.error(
                    "if word_embedding_type is fasttext and word_ngrams is 1, "
                    "binary should be set to True.")
                sys.exit(0)
            # not relevant to prune fasttext subword model
            self.model.save(np2vec_model_file)
        else:
            # prune non NP terms
            if self.prune_non_np:
                logger.info('pruning np2vec model')
                total_vec = 0
                vector_size = self.model.vector_size
                for word in self.model.wv.vocab.keys():
                    if self.is_marked(word) and len(word) > 1:
                        total_vec += 1
                logger.info(
                    "storing %sx%s projection weights for NP's into %s",
                    total_vec, vector_size, np2vec_model_file)
                with utils.smart_open(np2vec_model_file, 'wb') as fout:
                    fout.write(utils.to_utf8("%s %s\n" % (total_vec, vector_size)))
                    # store NP vectors in sorted order: most frequent NP's at the top
                    for word, vocab in sorted(
                            iteritems(
                                self.model.wv.vocab), key=lambda item: -item[1].count):
                        if self.is_marked(word) and len(word) > 1:  # discard empty marked np's
                            embedding_vec = self.model.wv.syn0[vocab.index]
                            if binary:
                                fout.write(
                                    utils.to_utf8(word) + b" " + embedding_vec.tostring())
                            else:
                                fout.write(
                                    utils.to_utf8(
                                        "%s %s\n" %
                                        (word, ' '.join(
                                            "%f" %
                                            val for val in embedding_vec))))
                if not word2vec_format:
                    # pylint: disable=attribute-defined-outside-init
                    self.model = KeyedVectors.load_word2vec_format(np2vec_model_file,
                                                                   binary=binary)
            if not word2vec_format:
                self.model.save(np2vec_model_file)

    @classmethod
    def load(cls, np2vec_model_file, binary=False, word_ngrams=0, word2vec_format=True):
        """
        Load the np2vec model.

        Args:
            np2vec_model_file (str): the file containing the np2vec model to load
            binary (bool): boolean indicating whether the np2vec model to load is in binary format
            word_ngrams (int {1,0}): If 1, np2vec model to load uses word vectors with subword (
            ngrams) information.
            word2vec_format(bool): boolean indicating whether the model to load has been stored in
            original word2vec format.

        Returns:
            np2vec model to load
        """
        if word_ngrams == 0:
            if word2vec_format:
                return KeyedVectors.load_word2vec_format(np2vec_model_file, binary=binary)
            return KeyedVectors.load(np2vec_model_file, mmap='r')
        if word_ngrams == 1:
            return FastText.load(np2vec_model_file)
        logger.error('invalid value for \'word_ngrams\'')
        return None



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# Changes Made from original:
#   import paths
# ******************************************************************************
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utility to handle vocabularies."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import codecs
import os
import tensorflow as tf

# pylint: disable=no-name-in-module
from tensorflow.python.ops import lookup_ops

from nlp_architect.models.gnmt.utils import misc_utils as utils

# word level special token
UNK = "<unk>"
SOS = "<s>"
EOS = "</s>"
UNK_ID = 0

# char ids 0-255 come from utf-8 encoding bytes
# assign 256-300 to special chars
BOS_CHAR_ID = 256  # <begin sentence>
EOS_CHAR_ID = 257  # <end sentence>
BOW_CHAR_ID = 258  # <begin word>
EOW_CHAR_ID = 259  # <end word>
PAD_CHAR_ID = 260  # <padding>

DEFAULT_CHAR_MAXLEN = 50  # max number of chars for each word.


def _string_to_bytes(text, max_length):
    """Given string and length, convert to byte seq of at most max_length.

  This process mimics docqa/elmo's preprocessing:
  https://github.com/allenai/document-qa/blob/master/docqa/elmo/data.py

  Note that we make use of BOS_CHAR_ID and EOS_CHAR_ID in iterator_utils.py &
  our usage differs from docqa/elmo.

  Args:
    text: tf.string tensor of shape []
    max_length: max number of chars for each word.

  Returns:
    A tf.int32 tensor of the byte encoded text.
  """
    byte_ids = tf.to_int32(tf.decode_raw(text, tf.uint8))
    byte_ids = byte_ids[:max_length - 2]
    padding = tf.fill([max_length - tf.shape(byte_ids)[0] - 2], PAD_CHAR_ID)
    byte_ids = tf.concat(
        [[BOW_CHAR_ID], byte_ids, [EOW_CHAR_ID], padding], axis=0)
    tf.logging.info(byte_ids)

    byte_ids = tf.reshape(byte_ids, [max_length])
    tf.logging.info(byte_ids.get_shape().as_list())
    return byte_ids + 1


def tokens_to_bytes(tokens):
    """Given a sequence of strings, map to sequence of bytes.

  Args:
    tokens: A tf.string tensor

  Returns:
    A tensor of shape words.shape + [bytes_per_word] containing byte versions
    of each word.
  """
    bytes_per_word = DEFAULT_CHAR_MAXLEN
    with tf.device("/cpu:0"):
        tf.assert_rank(tokens, 1)
        shape = tf.shape(tokens)
        tf.logging.info(tokens)
        tokens_flat = tf.reshape(tokens, [-1])
        as_bytes_flat = tf.map_fn(
            fn=lambda x: _string_to_bytes(x, max_length=bytes_per_word),
            elems=tokens_flat,
            dtype=tf.int32,
            back_prop=False)
        tf.logging.info(as_bytes_flat)
        as_bytes = tf.reshape(as_bytes_flat, [shape[0], bytes_per_word])
    return as_bytes


def load_vocab(vocab_file):
    vocab = []
    with codecs.getreader("utf-8")(tf.gfile.GFile(vocab_file, "rb")) as f:
        vocab_size = 0
        for word in f:
            vocab_size += 1
            vocab.append(word.strip())
    return vocab, vocab_size


def check_vocab(vocab_file, out_dir, check_special_token=True, sos=None,
                eos=None, unk=None):
    """Check if vocab_file doesn't exist, create from corpus_file."""
    if tf.gfile.Exists(vocab_file):
        utils.print_out("# Vocab file %s exists" % vocab_file)
        vocab, vocab_size = load_vocab(vocab_file)
        if check_special_token:
            # Verify if the vocab starts with unk, sos, eos
            # If not, prepend those tokens & generate a new vocab file
            if not unk:
                unk = UNK
            if not sos:
                sos = SOS
            if not eos:
                eos = EOS
            assert len(vocab) >= 3
            if vocab[0] != unk or vocab[1] != sos or vocab[2] != eos:
                utils.print_out("The first 3 vocab words [%s, %s, %s]"
                                " are not [%s, %s, %s]" %
                                (vocab[0], vocab[1], vocab[2], unk, sos, eos))
                vocab = [unk, sos, eos] + vocab
                vocab_size += 3
                new_vocab_file = os.path.join(out_dir, os.path.basename(vocab_file))
                with codecs.getwriter("utf-8")(
                        tf.gfile.GFile(new_vocab_file, "wb")) as f:
                    for word in vocab:
                        f.write("%s\n" % word)
                vocab_file = new_vocab_file
    else:
        raise ValueError("vocab_file '%s' does not exist." % vocab_file)

    vocab_size = len(vocab)
    return vocab_size, vocab_file


def create_vocab_tables(src_vocab_file, tgt_vocab_file, share_vocab):
    """Creates vocab tables for src_vocab_file and tgt_vocab_file."""
    src_vocab_table = lookup_ops.index_table_from_file(
        src_vocab_file, default_value=UNK_ID)
    if share_vocab:
        tgt_vocab_table = src_vocab_table
    else:
        tgt_vocab_table = lookup_ops.index_table_from_file(
            tgt_vocab_file, default_value=UNK_ID)
    return src_vocab_table, tgt_vocab_table


def load_embed_txt(embed_file):
    """Load embed_file into a python dictionary.

  Note: the embed_file should be a Glove/word2vec formatted txt file. Assuming
  Here is an exampe assuming embed_size=5:

  the -0.071549 0.093459 0.023738 -0.090339 0.056123
  to 0.57346 0.5417 -0.23477 -0.3624 0.4037
  and 0.20327 0.47348 0.050877 0.002103 0.060547

  For word2vec format, the first line will be: <num_words> <emb_size>.

  Args:
    embed_file: file path to the embedding file.
  Returns:
    a dictionary that maps word to vector, and the size of embedding dimensions.
  """
    emb_dict = dict()
    emb_size = None

    is_first_line = True
    with codecs.getreader("utf-8")(tf.gfile.GFile(embed_file, "rb")) as f:
        for line in f:
            tokens = line.rstrip().split(" ")
            if is_first_line:
                is_first_line = False
                if len(tokens) == 2:  # header line
                    emb_size = int(tokens[1])
                    continue
            word = tokens[0]
            vec = list(map(float, tokens[1:]))
            emb_dict[word] = vec
            if emb_size:
                if emb_size != len(vec):
                    utils.print_out(
                        "Ignoring %s since embeding size is inconsistent." % word)
                    del emb_dict[word]
            else:
                emb_size = len(vec)
    return emb_dict, emb_size



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Data loader for penn tree bank dataset
"""
import os
import sys
import numpy as np
import urllib.request

LICENSE_URL = {'PTB': "http://www.fit.vutbr.cz/~imikolov/rnnlm/",
               'WikiText-103': "https://einstein.ai/research/the-wikitext-long-term-dependency-"
                               "language-modeling-dataset"}

SOURCE_URL = {'PTB': "http://www.fit.vutbr.cz/~imikolov/rnnlm/simple-examples.tgz",
              'WikiText-103': "https://s3.amazonaws.com/research.metamind.io/wikitext/"
                              + "wikitext-103-v1.zip"}
FILENAME = {'PTB': "simple-examples", 'WikiText-103': "wikitext-103"}
EXTENSION = {'PTB': "tgz", 'WikiText-103': "zip"}
FILES = {'PTB': lambda x: "data/ptb." + x + ".txt",
         'WikiText-103': lambda x: "wiki." + x + ".tokens"}


class PTBDictionary:
    """
    Class for generating a dictionary of all words in the PTB corpus
    """
    def __init__(self, data_dir=os.path.expanduser('~/data'), dataset='WikiText-103'):
        """
        Initialize class
        Args:
            data_dir: str, location of data
            dataset: str, name of data corpus
        """
        self.data_dir = data_dir
        self.dataset = dataset
        self.filepath = os.path.join(data_dir, FILENAME[self.dataset])
        self._maybe_download(data_dir)

        self.word2idx = {}
        self.idx2word = []

        self.load_dictionary()
        print("Loaded dictionary of words of size {}".format(len(self.idx2word)))
        self.sos_symbol = self.word2idx['<sos>']
        self.eos_symbol = self.word2idx['<eos>']
        self.save_dictionary()

    def add_word(self, word):
        """
        Method for adding a single word to the dictionary
        Args:
            word: str, word to be added

        Returns:
            None
        """
        if word not in self.word2idx:
            self.idx2word.append(word)
            self.word2idx[word] = len(self.idx2word) - 1
        return self.word2idx[word]

    def load_dictionary(self):
        """
        Populate the corpus with words from train, test and valid splits of data
        Returns:
            None
        """
        for split_type in ["train", "test", "valid"]:
            path = os.path.join(self.data_dir, FILENAME[self.dataset],
                                FILES[self.dataset](split_type))
            # Add words to the dictionary
            with open(path, 'r') as fp:
                tokens = 0
                for line in fp:
                    words = ['<sos>'] + line.split() + ['<eos>']
                    tokens += len(words)
                    for word in words:
                        self.add_word(word)

    def save_dictionary(self):
        """
        Save dictionary to file
        Returns:
            None
        """
        with open(os.path.join(self.data_dir, "dictionary.txt"), "w") as fp:
            for k in self.word2idx:
                fp.write("%s,%d\n" % (k, self.word2idx[k]))

    def _maybe_download(self, work_directory):
        """
        This function downloads the corpus if its not already present
        Args:
            work_directory: str, location to download data to
        Returns:
            None
        """
        if not os.path.exists(self.filepath):
            print('{} was not found in the directory: {}, looking for compressed version'
                  .format(FILENAME[self.dataset], self.filepath))
            full_filepath = os.path.join(work_directory,
                                         FILENAME[self.dataset] + "." + EXTENSION[self.dataset])
            if not os.path.exists(full_filepath):
                print('Did not find data')
                print('PTB can be downloaded from http://www.fit.vutbr.cz/~imikolov/rnnlm/ \n'
                      'wikitext can be downloaded from'
                      ' https://einstein.ai/research/the-wikitext-long-term-dependency-language'
                      '-modeling-dataset')
                print('\nThe terms and conditions of the data set license apply. Intel does not '
                      'grant any rights to the data files or database\n')
                response = input('\nTo download data from {}, please enter YES: '
                                 .format(LICENSE_URL[self.dataset]))
                res = response.lower().strip()
                if res == "yes" or (len(res) == 1 and res == 'y'):
                    print("Downloading...")
                    self._download_data(work_directory)
                    self._uncompress_data(work_directory)
                else:
                    print('Download declined. Response received {} != YES|Y. '.format(res))
                    print('Please download the model manually from the links above '
                          'and place in directory: {}'.format(work_directory))
                    sys.exit()
            else:
                self._uncompress_data(work_directory)

    def _download_data(self, work_directory):
        """
        This function downloads the corpus
        Args:
            work_directory: str, location to download data to
        Returns:
            None
        """
        work_directory = os.path.abspath(work_directory)
        if not os.path.exists(work_directory):
            os.mkdir(work_directory)

        headers = {'User-Agent': 'Mozilla/5.0'}

        full_filepath = os.path.join(work_directory, FILENAME[self.dataset] + "."
                                     + EXTENSION[self.dataset])
        req = urllib.request.Request(SOURCE_URL[self.dataset], headers=headers)
        data_handle = urllib.request.urlopen(req)
        with open(full_filepath, "wb") as fp:
            fp.write(data_handle.read())
        print('Successfully downloaded data to {}'.format(full_filepath))

    def _uncompress_data(self, work_directory):
        full_filepath = os.path.join(work_directory,
                                     FILENAME[self.dataset] + "." + EXTENSION[self.dataset])
        if EXTENSION[self.dataset] == "tgz":
            import tarfile
            with tarfile.open(full_filepath, "r:gz") as tar:
                tar.extractall(path=work_directory)
        if EXTENSION[self.dataset] == "zip":
            import zipfile
            with zipfile.ZipFile(full_filepath, 'r') as zip_handle:
                zip_handle.extractall(work_directory)

        print('Successfully unzipped data to {}'.format(os.path.join(work_directory,
                                                                     FILENAME[self.dataset])))


class PTBDataLoader:
    """
    Class that defines data loader
    """
    def __init__(self, word_dict, seq_len=100, data_dir=os.path.expanduser('~/data'),
                 dataset='WikiText-103', batch_size=32, skip=30, split_type="train", loop=True):
        """
        Initialize class
        Args:
            word_dict: PTBDictionary object
            seq_len: int, sequence length of data
            data_dir: str, location of corpus data
            dataset: str, name of corpus
            batch_size: int, batch size
            skip: int, number of words to skip over while generating batches
            split_type: str, train/test/valid
            loop: boolean, whether or not to loop over data when it runs out
        """
        self.seq_len = seq_len
        self.dataset = dataset

        self.loop = loop
        self.skip = skip

        self.word2idx = word_dict.word2idx
        self.idx2word = word_dict.idx2word

        self.data = self.load_series(os.path.join(data_dir, FILENAME[self.dataset],
                                                  FILES[self.dataset](split_type)))
        self.random_index = np.random.permutation(np.arange(0, self.data.shape[0] - self.seq_len,
                                                            self.skip))
        self.n_train = self.random_index.shape[0]

        self.batch_size = batch_size
        self.sample_count = 0

    def __iter__(self):
        return self

    def __next__(self):
        return self.get_batch()

    def reset(self):
        """
        Resets the sample count to zero, re-shuffles data
        Returns:
            None
        """
        self.sample_count = 0
        self.random_index = np.random.permutation(np.arange(0, self.data.shape[0] - self.seq_len,
                                                            self.skip))

    def get_batch(self):
        """
        Get one batch of the data
        Returns:
            None
        """
        if self.sample_count + self.batch_size > self.n_train:
            if self.loop:
                self.reset()
            else:
                raise StopIteration("Ran out of data")

        batch_x = []
        batch_y = []
        for _ in range(self.batch_size):
            c_i = int(self.random_index[self.sample_count])
            batch_x.append(self.data[c_i:c_i + self.seq_len])
            batch_y.append(self.data[c_i + 1:c_i + self.seq_len + 1])
            self.sample_count += 1
        batch = (np.array(batch_x), np.array(batch_y))

        return batch

    def load_series(self, path):
        """
        Load all the data into an array
        Args:
            path: str, location of the input data file

        Returns:

        """
        # Tokenize file content
        with open(path, 'r') as fp:
            ids = []
            for line in fp:
                words = line.split() + ['<eos>']
                for word in words:
                    ids.append(self.word2idx[word])

        data = np.array(ids)

        return data

    def decode_line(self, tokens):
        """
        Decode a given line from index to word
        Args:
            tokens: List of indexes

        Returns:
            str, a sentence
        """
        return " ".join([self.idx2word[t] for t in tokens])



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from os import path
from pathlib import Path

LIBRARY_PATH = Path(path.realpath(__file__)).parent
LIBRARY_ROOT = LIBRARY_PATH.parent
LIBRARY_OUT = Path(Path.home()) / 'nlp-architect' / 'cache'
LIBRARY_DATASETS = LIBRARY_ROOT / 'datasets'



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import re


# This file contains adapted open sourced code, publicly available at:
# https://github.com/elikip/bist-parser/blob/master/bmstparser/src/utils.py

# Things that were changed from the original:
# 1) Added input validation
# 2) Updated function and object names to dyNet 2.0.2 and Python 3
# 3) Removed external embeddings option
# 4) Reformatted code and variable names to conform with PEP8
# 5) Added dict_to_obj()
# 6) Added option for train() to get ConllEntry input
# 7) Added legal header

NUMBER_REGEX = re.compile("[0-9]+|[0-9]+\\.[0-9]+|[0-9]+[0-9,]+")


class ConllEntry:
    def __init__(self, eid, form, lemma, pos, cpos, feats=None, parent_id=None, relation=None,
                 deps=None, misc=None):
        self.id = eid
        self.form = form
        self.norm = normalize(form)
        self.cpos = cpos.upper()
        self.pos = pos.upper()
        self.parent_id = parent_id
        self.relation = relation

        self.lemma = lemma
        self.feats = feats
        self.deps = deps
        self.misc = misc

        self.pred_parent_id = None
        self.pred_relation = None

        self.vec = None
        self.lstms = None

    def __str__(self):
        values = [str(self.id), self.form, self.lemma, self.cpos, self.pos, self.feats,
                  str(self.pred_parent_id) if self.pred_parent_id is not None else None,
                  self.pred_relation, self.deps, self.misc]
        return '\t'.join(['_' if v is None else v for v in values])


def normalize(word):
    return 'NUM' if NUMBER_REGEX.match(word) else word.lower()



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import os
import re
import string
from typing import List

from nlp_architect.utils.io import load_json_file
from nlp_architect.utils.text import SpacyInstance

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))

STOP_WORDS_FILE = os.path.join(CURRENT_DIR, 'resources/stop_words_en.json')
PRONOUN_FILE = os.path.join(CURRENT_DIR, 'resources/pronoun_en.json')
PREPOSITION_FILE = os.path.join(CURRENT_DIR, 'resources/preposition_en.json')
DETERMINERS_FILE = os.path.join(CURRENT_DIR, 'resources/determiners_en.json')

DISAMBIGUATION_CATEGORY = ['disambig', 'disambiguation']


class StringUtils:
    spacy_no_parser = SpacyInstance(disable=['parser'])
    spacy_parser = SpacyInstance()
    stop_words = []
    pronouns = []
    preposition = []
    determiners = []

    def __init__(self):
        pass

    @staticmethod
    def is_stop(token: str) -> bool:
        if not StringUtils.stop_words:
            StringUtils.stop_words = load_json_file(STOP_WORDS_FILE)
            StringUtils.stop_words.extend(DISAMBIGUATION_CATEGORY)
        if token not in StringUtils.stop_words:
            return False
        return True

    @staticmethod
    def normalize_str(in_str: str) -> str:
        str_clean = re.sub('[' + string.punctuation + string.whitespace + ']', ' ',
                           in_str).strip().lower()
        if isinstance(str_clean, str):
            str_clean = str(str_clean)

        parser = StringUtils.spacy_no_parser.parser
        doc = parser(str_clean)
        ret_clean = []
        for token in doc:
            lemma = token.lemma_.strip()
            if not StringUtils.is_pronoun(lemma) and not StringUtils.is_stop(lemma):
                ret_clean.append(token.lemma_)

        return ' '.join(ret_clean)

    @staticmethod
    def is_pronoun(in_str: str) -> bool:
        if not StringUtils.pronouns:
            StringUtils.pronouns = load_json_file(PRONOUN_FILE)

        tokens = in_str.split()
        if len(tokens) == 1:
            if tokens[0] in StringUtils.pronouns:
                return True
        return False

    @staticmethod
    def is_determiner(in_str: str) -> bool:
        if not StringUtils.determiners:
            StringUtils.determiners = load_json_file(DETERMINERS_FILE)

        tokens = in_str.split()
        if len(tokens) == 1:
            if tokens[0] in StringUtils.determiners:
                return True
        return False

    @staticmethod
    def is_preposition(in_str: str) -> bool:
        if not StringUtils.preposition:
            StringUtils.preposition = load_json_file(PREPOSITION_FILE)

        tokens = in_str.split()
        if len(tokens) == 1:
            if tokens[0] in StringUtils.preposition:
                return True
        return False

    @staticmethod
    def normalize_string_list(str_list: str) -> List[str]:
        ret_list = []
        for _str in str_list:
            normalize_str = StringUtils.normalize_str(_str)
            if normalize_str != '':
                ret_list.append(normalize_str)
        return ret_list

    @staticmethod
    def find_head_lemma_pos_ner(x: str):
        """"

        :param x: mention
        :return: the head word and the head word lemma of the mention
        """
        head = "UNK"
        lemma = "UNK"
        pos = "UNK"
        ner = "UNK"

        # pylint: disable=not-callable
        doc = StringUtils.spacy_parser.parser(x)
        for tok in doc:
            if tok.head == tok:
                head = tok.text
                lemma = tok.lemma_
                pos = tok.pos_

        for ent in doc.ents:
            if ent.root.text == head:
                ner = ent.label_

        return head, lemma, pos, ner



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# Changes Made from original:
#   import paths
# ******************************************************************************
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# pylint: skip-file
"""For loading data into NMT models."""
from __future__ import print_function

import collections

import tensorflow as tf

from nlp_architect.models.gnmt.utils import vocab_utils


__all__ = ["BatchedInput", "get_iterator", "get_infer_iterator"]


# NOTE(ebrevdo): When we subclass this, instances' __dict__ becomes empty.
class BatchedInput(
    collections.namedtuple("BatchedInput",
                           ("initializer", "source", "target_input",
                            "target_output", "source_sequence_length",
                            "target_sequence_length"))):
    pass


def get_infer_iterator(src_dataset,
                       src_vocab_table,
                       batch_size,
                       eos,
                       src_max_len=None,
                       use_char_encode=False):
    if use_char_encode:
        src_eos_id = vocab_utils.EOS_CHAR_ID
    else:
        src_eos_id = tf.cast(src_vocab_table.lookup(tf.constant(eos)), tf.int32)
    src_dataset = src_dataset.map(lambda src: tf.string_split([src]).values)

    if src_max_len:
        src_dataset = src_dataset.map(lambda src: src[:src_max_len])

    if use_char_encode:
        # Convert the word strings to character ids
        src_dataset = src_dataset.map(
            lambda src: tf.reshape(vocab_utils.tokens_to_bytes(src), [-1]))
    else:
        # Convert the word strings to ids
        src_dataset = src_dataset.map(
            lambda src: tf.cast(src_vocab_table.lookup(src), tf.int32))

    # Add in the word counts.
    if use_char_encode:
        src_dataset = src_dataset.map(
            lambda src: (src,
                         tf.to_int32(
                             tf.size(src) / vocab_utils.DEFAULT_CHAR_MAXLEN)))
    else:
        src_dataset = src_dataset.map(lambda src: (src, tf.size(src)))

    def batching_func(x):
        return x.padded_batch(
            batch_size,
            # The entry is the source line rows;
            # this has unknown-length vectors.  The last entry is
            # the source row size; this is a scalar.
            padded_shapes=(
                tf.TensorShape([None]),  # src
                tf.TensorShape([])),  # src_len
            # Pad the source sequences with eos tokens.
            # (Though notice we don't generally need to do this since
            # later on we will be masking out calculations past the true sequence.
            padding_values=(
                src_eos_id,  # src
                0))  # src_len -- unused

    batched_dataset = batching_func(src_dataset)
    batched_iter = batched_dataset.make_initializable_iterator()
    (src_ids, src_seq_len) = batched_iter.get_next()
    return BatchedInput(
        initializer=batched_iter.initializer,
        source=src_ids,
        target_input=None,
        target_output=None,
        source_sequence_length=src_seq_len,
        target_sequence_length=None)


def get_iterator(src_dataset,
                 tgt_dataset,
                 src_vocab_table,
                 tgt_vocab_table,
                 batch_size,
                 sos,
                 eos,
                 random_seed,
                 num_buckets,
                 src_max_len=None,
                 tgt_max_len=None,
                 num_parallel_calls=4,
                 output_buffer_size=None,
                 skip_count=None,
                 num_shards=1,
                 shard_index=0,
                 reshuffle_each_iteration=True,
                 use_char_encode=False):
    if not output_buffer_size:
        output_buffer_size = batch_size * 1000

    if use_char_encode:
        src_eos_id = vocab_utils.EOS_CHAR_ID
    else:
        src_eos_id = tf.cast(src_vocab_table.lookup(tf.constant(eos)), tf.int32)

    tgt_sos_id = tf.cast(tgt_vocab_table.lookup(tf.constant(sos)), tf.int32)
    tgt_eos_id = tf.cast(tgt_vocab_table.lookup(tf.constant(eos)), tf.int32)

    src_tgt_dataset = tf.data.Dataset.zip((src_dataset, tgt_dataset))

    src_tgt_dataset = src_tgt_dataset.shard(num_shards, shard_index)
    if skip_count is not None:
        src_tgt_dataset = src_tgt_dataset.skip(skip_count)

    src_tgt_dataset = src_tgt_dataset.shuffle(
        output_buffer_size, random_seed, reshuffle_each_iteration)

    src_tgt_dataset = src_tgt_dataset.map(
        lambda src, tgt: (
            tf.string_split([src]).values, tf.string_split([tgt]).values),
        num_parallel_calls=num_parallel_calls).prefetch(output_buffer_size)

    # Filter zero length input sequences.
    src_tgt_dataset = src_tgt_dataset.filter(
        lambda src, tgt: tf.logical_and(tf.size(src) > 0, tf.size(tgt) > 0))

    if src_max_len:
        src_tgt_dataset = src_tgt_dataset.map(
            lambda src, tgt: (src[:src_max_len], tgt),
            num_parallel_calls=num_parallel_calls).prefetch(output_buffer_size)
    if tgt_max_len:
        src_tgt_dataset = src_tgt_dataset.map(
            lambda src, tgt: (src, tgt[:tgt_max_len]),
            num_parallel_calls=num_parallel_calls).prefetch(output_buffer_size)

    # Convert the word strings to ids.  Word strings that are not in the
    # vocab get the lookup table's default_value integer.
    if use_char_encode:
        src_tgt_dataset = src_tgt_dataset.map(
            lambda src, tgt: (tf.reshape(vocab_utils.tokens_to_bytes(src), [-1]),
                              tf.cast(tgt_vocab_table.lookup(tgt), tf.int32)),
            num_parallel_calls=num_parallel_calls)
    else:
        src_tgt_dataset = src_tgt_dataset.map(
            lambda src, tgt: (tf.cast(src_vocab_table.lookup(src), tf.int32),
                              tf.cast(tgt_vocab_table.lookup(tgt), tf.int32)),
            num_parallel_calls=num_parallel_calls)

    src_tgt_dataset = src_tgt_dataset.prefetch(output_buffer_size)
    # Create a tgt_input prefixed with <sos> and a tgt_output suffixed with <eos>.
    src_tgt_dataset = src_tgt_dataset.map(
        lambda src, tgt: (src,
                          tf.concat(([tgt_sos_id], tgt), 0),
                          tf.concat((tgt, [tgt_eos_id]), 0)),
        num_parallel_calls=num_parallel_calls).prefetch(output_buffer_size)
    # Add in sequence lengths.
    if use_char_encode:
        src_tgt_dataset = src_tgt_dataset.map(
            lambda src, tgt_in, tgt_out: (
                src, tgt_in, tgt_out,
                tf.to_int32(tf.size(src) / vocab_utils.DEFAULT_CHAR_MAXLEN),
                tf.size(tgt_in)),
            num_parallel_calls=num_parallel_calls)
    else:
        src_tgt_dataset = src_tgt_dataset.map(
            lambda src, tgt_in, tgt_out: (
                src, tgt_in, tgt_out, tf.size(src), tf.size(tgt_in)),
            num_parallel_calls=num_parallel_calls)

    src_tgt_dataset = src_tgt_dataset.prefetch(output_buffer_size)

    # Bucket by source sequence length (buckets for lengths 0-9, 10-19, ...)
    def batching_func(x):
        return x.padded_batch(
            batch_size,
            # The first three entries are the source and target line rows;
            # these have unknown-length vectors.  The last two entries are
            # the source and target row sizes; these are scalars.
            padded_shapes=(
                tf.TensorShape([None]),  # src
                tf.TensorShape([None]),  # tgt_input
                tf.TensorShape([None]),  # tgt_output
                tf.TensorShape([]),  # src_len
                tf.TensorShape([])),  # tgt_len
            # Pad the source and target sequences with eos tokens.
            # (Though notice we don't generally need to do this since
            # later on we will be masking out calculations past the true sequence.
            padding_values=(
                src_eos_id,  # src
                tgt_eos_id,  # tgt_input
                tgt_eos_id,  # tgt_output
                0,  # src_len -- unused
                0))  # tgt_len -- unused

    if num_buckets > 1:

        def key_func(unused_1, unused_2, unused_3, src_len, tgt_len):
            # Calculate bucket_width by maximum source sequence length.
            # Pairs with length [0, bucket_width) go to bucket 0, length
            # [bucket_width, 2 * bucket_width) go to bucket 1, etc.  Pairs with length
            # over ((num_bucket-1) * bucket_width) words all go into the last bucket.
            if src_max_len:
                bucket_width = (src_max_len + num_buckets - 1) // num_buckets
            else:
                bucket_width = 10

            # Bucket sentence pairs by the length of their source sentence and target
            # sentence.
            bucket_id = tf.maximum(src_len // bucket_width, tgt_len // bucket_width)
            return tf.to_int64(tf.minimum(num_buckets, bucket_id))

        def reduce_func(unused_key, windowed_data):
            return batching_func(windowed_data)

        batched_dataset = src_tgt_dataset.apply(
            tf.contrib.data.group_by_window(
                key_func=key_func, reduce_func=reduce_func, window_size=batch_size))

    else:
        batched_dataset = batching_func(src_tgt_dataset)
    batched_iter = batched_dataset.make_initializable_iterator()
    (src_ids, tgt_input_ids, tgt_output_ids, src_seq_len,
     tgt_seq_len) = (batched_iter.get_next())
    return BatchedInput(
        initializer=batched_iter.initializer,
        source=src_ids,
        target_input=tgt_input_ids,
        target_output=tgt_output_ids,
        source_sequence_length=src_seq_len,
        target_sequence_length=tgt_seq_len)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import argparse
import json
import logging

from nlp_architect.data.cdc_resources.relations.referent_dict_relation_extraction import \
    ReferentDictRelationExtraction
from nlp_architect.models.cross_doc_coref.system.cdc_utils import load_mentions_vocab_from_files
from nlp_architect.utils import io

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

parser = argparse.ArgumentParser(description='Create Referent dictionary dataset only dump')

parser.add_argument('--ref_dict', type=str, help='referent dictionary file', required=True)

parser.add_argument('--mentions', type=str, help='dataset mentions', required=True)

parser.add_argument('--output', type=str, help='location were to create dump file', required=True)

args = parser.parse_args()


def ref_dict_dump():
    logger.info('Extracting referent dict dump, this may take a while...')
    ref_dict_file = args.ref_dict
    out_file = args.output
    mentions_entity_gold_file = [args.mentions]
    vocab = load_mentions_vocab_from_files(mentions_entity_gold_file, True)

    ref_dict = ReferentDictRelationExtraction.load_reference_dict(ref_dict_file)

    ref_dict_for_vocab = {}
    for word in vocab:
        if word in ref_dict:
            ref_dict_for_vocab[word] = ref_dict[word]

    logger.info('Found %d words from vocabulary', len(ref_dict_for_vocab.keys()))
    logger.info('Preparing to save refDict output file')
    with open(out_file, 'w') as f:
        json.dump(ref_dict_for_vocab, f)
    logger.info('Done saved to-%s', out_file)


if __name__ == '__main__':
    io.validate_existing_filepath(args.mentions)
    io.validate_existing_filepath(args.output)
    io.validate_existing_filepath(args.ref_dict)
    ref_dict_dump()



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import re
import string
from typing import Set, Dict

from nlp_architect.utils.string_utils import StringUtils

PART_NAME_CATEGORIES = ['name', 'given name', 'surname']
DISAMBIGUATION_TITLE = '(disambiguation)'
DISAMBIGUATION_CATEGORY = ['disambig', 'disambiguation']


class WikipediaPageExtractedRelations(object):
    def __init__(self, is_part_name: bool = False, is_disambiguation: bool = False,
                 parenthesis: Set[str] = None,
                 disambiguation_links: Set[str] = None, categories: Set[str] = None,
                 aliases: Set[str] = None,
                 be_comp: Set[str] = None,
                 disambiguation_links_norm: Set[str] = None, categories_norm: Set[str] = None,
                 aliases_norm: Set[str] = None,
                 title_parenthesis_norm: Set[str] = None, be_comp_norm: Set[str] = None) -> None:
        """
        Object represent a Wikipedia Relations Schema

        Args:
            is_part_name (bool): Weather page title is part of a Name (ie-family name/given name..)
            is_disambiguation (bool): Weather page is a disambiguation page
            parenthesis (set): a set of all parenthesis links/titles
            disambiguation_links (set): a set of all disambiguation links/titles
            categories (set): a set of all category links/titles
            aliases (set): a set of all aliases links/titles
            be_comp (set): a set of all "is a" links/titles
            disambiguation_links_norm (set): same as disambiguation_link just normalized
            categories_norm (set): same as categories just normalized, lower and clean
            aliases_norm (set): same as aliases just normalized, lower and clean
            title_parenthesis_norm (set): same as parenthesis just normalized, lower and clean
            be_comp_norm (set): same as be_comp just normalized, lower and clean
        """
        self.is_part_name = is_part_name
        self.is_disambiguation = is_disambiguation
        self.disambiguation_links = disambiguation_links
        self.title_parenthesis = parenthesis
        self.categories = categories
        self.aliases = aliases
        self.be_comp = be_comp

        self.disambiguation_links_norm = disambiguation_links_norm
        self.categories_norm = categories_norm
        self.aliases_norm = aliases_norm
        self.title_parenthesis_norm = title_parenthesis_norm
        self.be_comp_norm = be_comp_norm

    def extract_relations_from_text_v0(self, text):
        self.disambiguation_links = set()
        self.categories = set()
        self.title_parenthesis = set()

        self.disambiguation_links_norm = set()
        self.categories_norm = set()
        self.title_parenthesis_norm = set()
        self.be_comp_norm = set()

        ext_links = set()
        title_parenthesis = set()

        text_lines = text.split('\n')
        for line in text_lines:
            cat_links = self.extract_categories(line)
            if not self.is_part_name:
                self.is_part_name = self.is_name_part(line)
                if not self.is_part_name and [s for s in PART_NAME_CATEGORIES if s in cat_links]:
                    self.is_part_name = True

            self.categories.update(cat_links)
            self.categories_norm.update(StringUtils.normalize_string_list(cat_links))

            links, parenthesis_links = self.extract_links_and_parenthesis(line)
            ext_links.update(links)
            title_parenthesis.update(parenthesis_links)

        if self.is_disambiguation:
            self.disambiguation_links = ext_links
            self.disambiguation_links_norm = StringUtils.normalize_string_list(ext_links)
            self.title_parenthesis = title_parenthesis
            self.title_parenthesis_norm = StringUtils.normalize_string_list(title_parenthesis)

    def __str__(self) -> str:
        return str(self.is_disambiguation) + ', ' + str(self.is_part_name) + ', ' + \
            str(self.disambiguation_links) + ', ' + str(self.be_comp) + ', ' + str(
            self.title_parenthesis) + ', ' + str(self.categories)

    def toJson(self) -> Dict:
        result_dict = dict()
        result_dict['isPartName'] = self.is_part_name
        result_dict['isDisambiguation'] = self.is_disambiguation

        if self.disambiguation_links is not None:
            result_dict['disambiguationLinks'] = list(self.disambiguation_links)
            result_dict['disambiguationLinksNorm'] = list(self.disambiguation_links_norm)
        if self.categories is not None:
            result_dict['categories'] = list(self.categories)
            result_dict['categoriesNorm'] = list(self.categories_norm)
        if self.aliases is not None:
            result_dict['aliases'] = list(self.aliases)
        if self.title_parenthesis is not None:
            result_dict['titleParenthesis'] = list(self.title_parenthesis)
            result_dict['titleParenthesisNorm'] = list(self.title_parenthesis_norm)
        if self.be_comp_norm is not None:
            result_dict['beCompRelations'] = list(self.be_comp)
            result_dict['beCompRelationsNorm'] = list(self.be_comp_norm)

        return result_dict

    @staticmethod
    def extract_categories(line: str) -> Set[str]:
        categories = set()
        category_form1 = re.findall(r'\[\[Category:(.*)\]\]', line)
        for cat in category_form1:
            if DISAMBIGUATION_TITLE in cat:
                cat = cat.replace(DISAMBIGUATION_TITLE, '')
            categories.add(cat)

        prog = re.search('^{{(disambig.*|Disambig.*)}}$', line)
        if prog is not None:
            category_form2 = prog.group(1)
            cats = category_form2.split('|')
            categories.update(cats)

        return categories

    @staticmethod
    def extract_links_and_parenthesis(line: str):
        links = set()
        parenthesis_links = set()
        ext_links = re.findall(r'\[\[(.*)\]\]', line)
        for link in ext_links:
            split_link = link.split('|')
            for s_link in split_link:
                parenthesis_clean = None
                matcher = re.match(r'(.*)\s?\((.*)\)', s_link)
                if matcher:
                    s_link = matcher.group(1)
                    parenthesis_match = matcher.group(2)
                    if parenthesis_match.lower() != 'disambiguation':
                        parenthesis_clean = re.sub(
                            '[' + string.punctuation + string.whitespace + ']', ' ',
                            parenthesis_match).strip()

                s_link_clean = re.sub('[' + string.punctuation + string.whitespace + ']', ' ',
                                      s_link).strip()
                if parenthesis_clean is not None and DISAMBIGUATION_TITLE not in parenthesis_clean:
                    parenthesis_links.add(parenthesis_clean)

                links.add(s_link_clean)

        return links, parenthesis_links

    @staticmethod
    def is_name_part(line: str) -> bool:
        line = line.lower()
        val = False
        if WikipediaPageExtractedRelations.find_in_line(line, '===as surname==='):
            val = True
        elif WikipediaPageExtractedRelations.find_in_line(line, '===as given name==='):
            val = True
        elif WikipediaPageExtractedRelations.find_in_line(line, '===given names==='):
            val = True
        elif WikipediaPageExtractedRelations.find_in_line(line, '==as a surname=='):
            val = True
        elif WikipediaPageExtractedRelations.find_in_line(line, '==people with the surname=='):
            val = True
        elif WikipediaPageExtractedRelations.find_in_line(line, '==family name and surname=='):
            val = True
        elif WikipediaPageExtractedRelations.find_in_line(line, 'category:given names'):
            val = True
        elif WikipediaPageExtractedRelations.find_in_line(line, '{{given name}}'):
            val = True
        return val

    @staticmethod
    def find_in_line(text: str, pattern: str) -> bool:
        found = re.findall(pattern, text)
        if found:
            return True
        return False



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ****************************************************************************
import tensorflow as tf


class MostCommonWordSense(object):
    def __init__(self, epochs, batch_size, callback_args=None):
        self.optimizer = tf.keras.optimizers.SGD(lr=0.1, decay=1e-6, momentum=0.9, nesterov=True)
        self.loss = 'mean_squared_error'
        self.epochs = epochs
        self.batch_size = batch_size
        self.model = None
        self.callback_args = callback_args

    def build(self, input_dim):
        # setup model layers
        model = tf.keras.models.Sequential()
        model.add(tf.keras.layers.Dense(100, activation='relu', input_dim=input_dim))
        model.add(tf.keras.layers.Dropout(0.5))
        model.add(tf.keras.layers.Dense(2, activation='softmax'))
        model.compile(loss=self.loss, optimizer=self.optimizer)
        self.model = model

    def fit(self, train_set):
        self.model.fit(train_set['X'], train_set['y'], epochs=self.epochs,
                       batch_size=self.batch_size)

    def save(self, save_path):
        self.model.save(save_path)

    def load(self, model_path):
        self.model = tf.keras.models.load_model(model_path)

    def eval(self, valid_set):
        eval_rate = self.model.evaluate(valid_set['X'], valid_set['y'], batch_size=self.batch_size)
        return eval_rate

    def get_outputs(self, valid_set):
        return self.model.predict(valid_set)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import nltk
from nltk.corpus import wordnet as wn

from nlp_architect.data.cdc_resources.data_types.wn.wordnet_page import WordnetPage
from nlp_architect.utils.string_utils import StringUtils


class WordnetOnline(object):
    def __init__(self):
        self.cache = dict()
        nltk.download('wordnet')

    def get_pages(self, mention):
        if mention.tokens_str in self.cache:
            return self.cache[mention.tokens_str]

        head_synonyms, head_names_derivationally = self.extract_synonyms_and_derivation(
            mention.mention_head)
        head_lemma_synonyms, head_lemma_derivationally = self.extract_synonyms_and_derivation(
            mention.mention_head_lemma)
        clean_phrase = StringUtils.normalize_str(mention.tokens_str)
        all_clean_words_synonyms = self.all_clean_words_synonyms(clean_phrase)

        wordnet_page = WordnetPage(mention.tokens_str, clean_phrase, mention.mention_head,
                                   mention.mention_head_lemma,
                                   head_synonyms,
                                   head_lemma_synonyms, head_names_derivationally,
                                   head_lemma_derivationally,
                                   all_clean_words_synonyms)

        self.cache[mention.tokens_str] = wordnet_page
        return wordnet_page

    @staticmethod
    def extract_synonyms_and_derivation(word):
        lemma_names = set()
        derivationally_related_forms = set()
        for synset in wn.synsets(word):
            for lemma in synset.lemmas():
                lemma_name = lemma.name().replace('_', ' ')
                if not StringUtils.is_stop(lemma_name.lower()):
                    lemma_names.add(lemma_name)

                derivationally_related_forms.update(
                    [l.name().replace('_', ' ') for l in lemma.derivationally_related_forms()
                     if not StringUtils.is_stop(l.name().lower())])

        return lemma_names, derivationally_related_forms

    @staticmethod
    def all_clean_words_synonyms(clean_phrase):
        words = clean_phrase.split()
        return [set([lemma.lower().replace('_', ' ')
                     for synset in wn.synsets(w)
                     for lemma in synset.lemma_names() if not StringUtils.is_stop(lemma.lower())])
                for w in words]



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

from __future__ import absolute_import, division, print_function, unicode_literals

import json
import os
import sys

import numpy as np
from nlp_architect.utils.generic import pad_sentences
from nlp_architect.utils.text import SpacyInstance, Vocabulary, character_vector_generator, \
    word_vector_generator


class IntentDataset(object):
    """
    Intent extraction dataset base class

    Args:
        sentence_length (int): max sentence length
    """

    def __init__(self, sentence_length=50, word_length=12):
        self.data_dict = {}
        self.vecs = {}
        self.sentence_len = sentence_length
        self.word_len = word_length

        self._tokens_vocab = Vocabulary(2)
        self._chars_vocab = Vocabulary(2)
        self._tags_vocab = Vocabulary(1)
        self._intents_vocab = Vocabulary()

    def _load_data(self, train_set, test_set):
        # vectorize
        # add offset of 2 for PAD and OOV
        train_size = len(train_set)
        test_size = len(test_set)
        texts, tags, intents = list(zip(*train_set + test_set))
        text_vectors, self._tokens_vocab = word_vector_generator(texts, lower=True, start=2)
        tag_vectors, self._tags_vocab = word_vector_generator(tags, lower=False, start=1)
        chars_vectors, self._chars_vocab = character_vector_generator(texts, start=2)
        i, self._intents_vocab = word_vector_generator([intents])
        i = np.asarray(i[0])

        text_vectors = pad_sentences(text_vectors, max_length=self.sentence_len)
        tag_vectors = pad_sentences(tag_vectors, max_length=self.sentence_len)
        chars_vectors = [pad_sentences(d, max_length=self.word_len) for d in chars_vectors]
        zeros = np.zeros((len(chars_vectors), self.sentence_len, self.word_len))
        for idx, d in enumerate(chars_vectors):
            d = d[:self.sentence_len]
            zeros[idx, :d.shape[0]] = d
        chars_vectors = zeros.astype(dtype=np.int32)

        self.vecs['train'] = [text_vectors[:train_size],
                              chars_vectors[:train_size],
                              i[:train_size],
                              tag_vectors[:train_size]]
        self.vecs['test'] = [text_vectors[-test_size:],
                             chars_vectors[-test_size:],
                             i[-test_size:],
                             tag_vectors[-test_size:]]

    @property
    def word_vocab_size(self):
        """int: vocabulary size"""
        return len(self._tokens_vocab) + 2

    @property
    def char_vocab_size(self):
        """int: char vocabulary size"""
        return len(self._chars_vocab) + 2

    @property
    def label_vocab_size(self):
        """int: label vocabulary size"""
        return len(self._tags_vocab) + 1

    @property
    def intent_size(self):
        """int: intent label vocabulary size"""
        return len(self._intents_vocab)

    @property
    def word_vocab(self):
        """dict: tokens vocabulary"""
        return self._tokens_vocab

    @property
    def char_vocab(self):
        """dict: word character vocabulary"""
        return self._chars_vocab

    @property
    def tags_vocab(self):
        """dict: labels vocabulary"""
        return self._tags_vocab

    @property
    def intents_vocab(self):
        """dict: intent labels vocabulary"""
        return self._intents_vocab

    @property
    def train_set(self):
        """:obj:`tuple` of :obj:`numpy.ndarray`: train set"""
        return self.vecs['train']

    @property
    def test_set(self):
        """:obj:`tuple` of :obj:`numpy.ndarray`: test set"""
        return self.vecs['test']


class TabularIntentDataset(IntentDataset):
    """
    Tabular Intent/Slot tags dataset loader.
    Compatible with many sequence tagging datasets (ATIS, CoNLL, etc..)
    data format must be int tabular format where:
    - one word per line with tag annotation and intent type separated
    by tabs <token>\t<tag_label>\t<intent>\n
    - sentences are separated by an empty line

    Args:
        train_file (str): path to train set file
        test_file (str): path to test set file
        sentence_length (int): max sentence length
        word_length (int): max word length
    """
    files = ['train', 'test']

    def __init__(self, train_file, test_file, sentence_length=30, word_length=12):
        train_set_raw, test_set_raw = self._load_dataset(train_file, test_file)
        super(TabularIntentDataset, self).__init__(sentence_length=sentence_length,
                                                   word_length=word_length)

        self._load_data(train_set_raw, test_set_raw)

    def _load_dataset(self, train_file, test_file):
        """returns a tuple of train/test with 3-tuple of tokens, tags, intent_type"""
        train = self._parse_sentences(self._read_file(train_file))
        test = self._parse_sentences(self._read_file(test_file))
        return train, test

    def _read_file(self, path):
        with open(path, encoding='utf-8', errors='ignore') as fp:
            data = fp.readlines()
        return self._split_into_sentences(data)

    @staticmethod
    def _split_into_sentences(file_lines):
        sents = []
        s = []
        for line in file_lines:
            line = line.strip()
            if not line:
                sents.append(s)
                s = []
                continue
            s.append(line)
        if s:
            sents.append(s)
        return sents

    @staticmethod
    def _parse_sentences(sentences):
        encoded_sentences = []
        for sen in sentences:
            tokens = []
            tags = []
            intent = None
            for line in sen:
                t, s, i = line.split()
                tokens.append(t)
                tags.append(s)
                intent = i
                if intent is None:
                    intent = i
            encoded_sentences.append((tokens, tags, intent))
        return encoded_sentences


class SNIPS(IntentDataset):
    """
    SNIPS dataset class

    Args:
            path (str): dataset path
            sentence_length (int, optional): max sentence length
            word_length (int, optional): max word length
    """
    train_files = [
        'AddToPlaylist/train_AddToPlaylist_full.json',
        'BookRestaurant/train_BookRestaurant_full.json',
        'GetWeather/train_GetWeather_full.json',
        'PlayMusic/train_PlayMusic_full.json',
        'RateBook/train_RateBook_full.json',
        'SearchCreativeWork/train_SearchCreativeWork_full.json',
        'SearchScreeningEvent/train_SearchScreeningEvent_full.json'
    ]
    test_files = [
        'AddToPlaylist/validate_AddToPlaylist.json',
        'BookRestaurant/validate_BookRestaurant.json',
        'GetWeather/validate_GetWeather.json',
        'PlayMusic/validate_PlayMusic.json',
        'RateBook/validate_RateBook.json',
        'SearchCreativeWork/validate_SearchCreativeWork.json',
        'SearchScreeningEvent/validate_SearchScreeningEvent.json'
    ]
    files = ['train', 'test']

    def __init__(self, path, sentence_length=30, word_length=12):
        if path is None or not os.path.isdir(path):
            print('invalid path for SNIPS dataset loader')
            sys.exit(0)
        self.dataset_root = path
        train_set_raw, test_set_raw = self._load_dataset()
        super(SNIPS, self).__init__(sentence_length=sentence_length,
                                    word_length=word_length)
        self._load_data(train_set_raw, test_set_raw)

    def _load_dataset(self):
        """returns a tuple of train/test with 3-tuple of tokens, tags, intent_type"""
        train_data = self._load_intents(self.train_files)
        test_data = self._load_intents(self.test_files)
        train = [(t, l, i) for i in sorted(train_data) for t, l in train_data[i]]
        test = [(t, l, i) for i in sorted(test_data) for t, l in test_data[i]]
        return train, test

    def _load_intents(self, files):
        data = {}
        for f in sorted(files):
            fname = os.path.join(self.dataset_root, f)
            intent = f.split(os.sep)[0]
            with open(fname, encoding='utf-8', errors='ignore') as fp:
                fdata = json.load(fp)
            entries = self._parse_json([d['data'] for d in fdata[intent]])
            data[intent] = entries
        return data

    def _parse_json(self, data):
        tok = SpacyInstance(disable=['tagger', 'ner', 'parser', 'vectors', 'textcat'])
        sentences = []
        for s in data:
            tokens = []
            tags = []
            for t in s:
                new_tokens = tok.tokenize(t['text'].strip())
                tokens += new_tokens
                ent = t.get('entity', None)
                if ent is not None:
                    tags += self._create_tags(ent, len(new_tokens))
                else:
                    tags += ['O'] * len(new_tokens)
            sentences.append((tokens, tags))
        return sentences

    @staticmethod
    def _create_tags(tag, length):
        labels = ['B-' + tag]
        if length > 1:
            for _ in range(length - 1):
                labels.append('I-' + tag)
        return labels



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import logging
import sys
from os import path

from argparse import ArgumentParser

from nlp_architect.models.np2vec import NP2vec
from nlp_architect.utils.io import validate_existing_filepath, check_size, load_json_file

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)


class SetExpand(object):
    """
        Set expansion module, given a trained np2vec model.
    """

    def __init__(self, np2vec_model_file, binary=False, word_ngrams=False, grouping=False,
                 light_grouping=False, grouping_map_dir=None):
        """
        Load the np2vec model for set expansion.

        Args:
            np2vec_model_file (str): the file containing the np2vec model to load
            binary (bool): boolean indicating whether the np2vec model to load is in binary format
            word_ngrams (int {1,0}): If 1, np2vec model to load uses word vectors with subword (
            ngrams) information.
            light_grouping (bool): boolean indicating whether to load all maps for grouping.
            grouping_map_dir (str): path to the directory containing maps for grouping.
        Returns:
            np2vec model to load
        """
        self.grouping = grouping
        if grouping:
            # load grouping info
            logger.info('loading grouping data')
            if not grouping_map_dir:
                grouping_map_dir = path.dirname(np2vec_model_file)
            self.np2id = load_json_file(path.join(grouping_map_dir, 'np2id'))
            if not light_grouping:
                self.id2rep = load_json_file(path.join(grouping_map_dir, 'id2rep'))
                self.id2group = load_json_file(path.join(grouping_map_dir, 'id2group'))
        logger.info('loadind model...')
        self.np2vec_model = NP2vec.load(np2vec_model_file, binary=binary, word_ngrams=word_ngrams)
        # extract the first term of the model in order to get the marking character
        logger.info('compute L2 norm')
        first_term = next(iter(self.np2vec_model.vocab.keys()))
        self.mark_char = first_term[-1]
        # Precompute L2-normalized vectors.
        self.np2vec_model.init_sims()
        logger.info('done init')

    def term2id(self, term, suffix=True):
        """
        Given an term, return its id.

        Args:
            term(str): term (noun phrase)

        Returns:
            its id (if is part of the model)
        """
        if self.grouping:
            if term not in self.np2id.keys():
                return None
            term = self.np2id[term]
        id = term.replace(' ', self.mark_char)
        if suffix:
            id += self.mark_char
        if id not in self.np2vec_model.vocab:
            return None
        return id

    def __id2term(self, id):
        """
        Given the id of a noun phrase, return the noun phrase string.

        Args:
            id(str): id

         Returns:
            term (noun phrase)
        """
        norm = id.replace(self.mark_char, ' ')[:-1]
        if self.grouping:
            if norm in self.id2rep:
                return self.id2rep[norm]
            logger.warning("id:#%s#, norm:#%s# is not in id2rep")
            return ""
        return norm

    def get_vocab(self):
        """
        Return the vocabulary as the list of terms.

        Returns:
            the list of terms.
        """
        vocab = []
        for id in self.np2vec_model.vocab:
            term = self.__id2term(id)
            if term is not None:
                vocab.append(term)
        return vocab

    def in_vocab(self, term):
        id = self.term2id(term)
        if id is None:
            return False
        return True

    def get_group(self, term):
        logger.info("get group of: %s", term)
        group = []
        if term in self.np2id:
            id = self.np2id[term]
            group = self.id2group[id]
        return group

    def similarity(self, terms, seed, threshold):

        similar = []
        seed_id = self.get_seed_id(seed)
        for term in terms:
            term_id = self.term2id(term)
            if term_id is not None:
                if self.seed2term_similarity(seed_id, [term_id]) > threshold:
                    similar.append(term)
            else:
                logger.info("term: %s is not in vocab", term)
        return similar

    # pylint: disable-msg=too-many-branches
    def expand(self, seed, topn=500):
        """
        Given a seed of terms, return the expanded set of terms.

        Args:
            seed: seed terms
            topn: maximal number of expanded terms to return

        Returns:
            up to topn expanded terms and their probabilities
        """
        seed_ids = list()
        upper = True
        lower = True
        for np in seed:
            np = np.strip()
            if not self.grouping and (upper or lower):
                # case feature is relevant only if we don't have grouping
                if np[0].islower():
                    upper = False
                else:
                    lower = False
            id = self.term2id(np)
            if id is not None:
                seed_ids.append(id)
            else:
                logger.warning("The term: '%s' is out-of-vocabulary.", np)
        if len(seed_ids) > 0:
            if not self.grouping and (upper or lower):
                res_id = self.np2vec_model.most_similar(seed_ids, topn=2 * topn)
            else:
                res_id = self.np2vec_model.most_similar(seed_ids, topn=topn)
            res = list()
            for r in res_id:
                if len(res) == topn:
                    break
                # pylint: disable=R0916
                if self.grouping or (not lower and not upper) or (upper and r[0][0].isupper()) or \
                        (lower and r[0][0].islower()):
                    term = self.__id2term(r[0])
                    if term is not None:
                        res.append((self.__id2term(r[0]), r[1]))
            ret_val = res
        else:
            logger.info("All the seed terms are out-of-vocabulary.")
            ret_val = None
        return ret_val

    def get_seed_id(self, seed):
        seed_ids = list()
        for np in seed:
            np = np.strip()
            id = self.term2id(np)
            if id is not None:
                seed_ids.append(id)
            else:
                logger.warning("The term: '%s' is out-of-vocabulary.", np)
        return seed_ids

    def term2term_similarity(self, term_id_1, term_id_2):
        """
        Compute cosine similarity between two term id's.
        Args:
            term_id_1: first term id
            term_id_2: second term id

        Returns:
            Similarity between the first and second term id's

        """
        logger.info('calculate similarity for: %s , %s', term_id_1, term_id_2)
        res = self.np2vec_model.similarity(term_id_1, term_id_2)
        logger.info("similarity result: %s", str(res))
        return res

    def seed2term_similarity(self, seed_id, term_id):
        """
        Compute cosine similarity between a seed terms and a term.
        Args:
            seed_id: seed term id's
            term_id: the term id

        Returns:
            Similarity between the seed terms and the term

        """
        logger.info('calculate similarity for: %s , %s', str(seed_id), term_id)
        res = self.np2vec_model.n_similarity(seed_id, list(term_id))
        logger.info("similarity result: %s", str(res))
        return res


if __name__ == "__main__":
    arg_parser = ArgumentParser(__doc__)
    arg_parser.add_argument(
        '--np2vec_model_file',
        help='path to the file with the np2vec model to load.',
        type=validate_existing_filepath)
    arg_parser.add_argument(
        '--binary',
        help='boolean indicating whether the model to load has been stored in binary format.',
        action='store_true')
    arg_parser.add_argument(
        '--word_ngrams',
        default=0,
        type=int,
        choices=[0, 1],
        help='If 0, the model to load stores word information. If 1, the model to load stores '
             'subword (ngrams) information; note that subword information is relevant only to '
             'fasttext models.')
    arg_parser.add_argument(
        '--topn',
        default=500,
        type=int,
        action=check_size(min_size=1),
        help='maximal number of expanded terms to return')
    arg_parser.add_argument(
        '--grouping',
        action='store_true',
        default=False,
        help='grouping mode')

    args = arg_parser.parse_args()

    se = SetExpand(np2vec_model_file=args.np2vec_model_file, binary=args.binary,
                   word_ngrams=args.word_ngrams, grouping=args.grouping)
    enter_seed_str = 'Enter the seed (comma-separated seed terms):'
    logger.info(enter_seed_str)
    for seed_str in sys.stdin:
        seed_list = seed_str.strip().split(',')
        exp = se.expand(seed_list, args.topn)
        logger.info('Expanded results:')
        logger.info(exp)
        logger.info(enter_seed_str)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import argparse
import gzip
import io
import json
import os
import posixpath
import re
import sys
import zipfile
from os import PathLike, makedirs, remove
from pathlib import Path
from urllib.parse import urlparse

import requests
from tqdm import tqdm

from nlp_architect.utils.generic import license_prompt


def download_unlicensed_file(url, sourcefile, destfile, totalsz=None):
    """
    Download the file specified by the given URL.

    Args:
        url (str): url to download from
        sourcefile (str): file to download from url
        destfile (str): save path
        totalsz (:obj:`int`, optional): total size of file
    """
    req = requests.get(posixpath.join(url, sourcefile),
                       stream=True)

    chunksz = 1024 ** 2
    if totalsz is None:
        if "Content-length" in req.headers:
            totalsz = int(req.headers["Content-length"])
            nchunks = totalsz // chunksz
        else:
            print("Unable to determine total file size.")
            nchunks = None
    else:
        nchunks = totalsz // chunksz

    print("Downloading file to: {}".format(destfile))
    with open(destfile, 'wb') as f:
        for data in tqdm(req.iter_content(chunksz), total=nchunks, unit="MB", file=sys.stdout):
            f.write(data)
    print("Download Complete")


def uncompress_file(filepath: str or os.PathLike, outpath='.'):
    """
    Unzip a file to the same location of filepath
    uses decompressing algorithm by file extension

    Args:
        filepath (str): path to file
        outpath (str): path to extract to
    """
    filepath = str(filepath)
    if filepath.endswith('.zip'):
        with zipfile.ZipFile(filepath) as z:
            z.extractall(outpath)
    elif filepath.endswith('.gz'):
        if os.path.isdir(outpath):
            raise ValueError('output path for gzip must be a file')
        with gzip.open(filepath, 'rb') as fp:
            file_content = fp.read()
        with open(outpath, 'wb') as fp:
            fp.write(file_content)
    else:
        raise ValueError('Unsupported archive provided. Method supports only .zip/.gz files.')


def gzip_str(g_str):
    """
    Transform string to GZIP coding

    Args:
        g_str (str): string of data

    Returns:
        GZIP bytes data
    """
    compressed_str = io.BytesIO()
    with gzip.GzipFile(fileobj=compressed_str, mode='w') as file_out:
        file_out.write((json.dumps(g_str).encode()))
    bytes_obj = compressed_str.getvalue()
    return bytes_obj


def check_directory_and_create(dir_path):
    """
    Check if given directory exists, create if not.

    Args:
        dir_path (str): path to directory
    """
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)


def walk_directory(directory, verbose=False):
    """Iterates a directory's text files and their contents."""
    for dir_path, _, filenames in os.walk(directory):
        for filename in filenames:
            file_path = os.path.join(dir_path, filename)
            if os.path.isfile(file_path) and not filename.startswith('.'):
                with io.open(file_path, 'r', encoding='utf-8') as file:
                    if verbose:
                        print('Reading {}'.format(filename))
                    doc_text = file.read()
                    yield filename, doc_text


def validate(*args):
    """
    Validate all arguments are of correct type and in correct range.
    Args:
        *args (tuple of tuples): Each tuple represents an argument validation like so:
        Option 1 - With range check:
            (arg, class, min_val, max_val)
        Option 2 - Without range check:
            (arg, class)
        If class is a tuple of type objects check if arg is an instance of any of the types.
        To allow a None valued argument, include type(None) in class.
        To disable lower or upper bound check, set min_val or max_val to None, respectively.
        If arg has the len attribute (such as string), range checks are performed on its length.
    """
    for arg in args:
        arg_val = arg[0]
        arg_type = (arg[1],) if isinstance(arg[1], type) else arg[1]
        if not isinstance(arg_val, arg_type):
            raise TypeError('Expected type {}'.format(' or '.join([t.__name__ for t in arg_type])))
        if arg_val is not None and len(arg) >= 4:
            name = 'of ' + arg[4] if len(arg) == 5 else ''
            arg_min = arg[2]
            arg_max = arg[3]
            if hasattr(arg_val, '__len__'):
                val = 'Length'
                num = len(arg_val)
            else:
                val = 'Value'
                num = arg_val
            if arg_min is not None and num < arg_min:
                raise ValueError('{} {} must be greater or equal to {}'.format(val, name, arg_min))
            if arg_max is not None and num >= arg_max:
                raise ValueError('{} {} must be less than {}'.format(val, name, arg_max))


def validate_existing_filepath(arg):
    """Validates an input argument is a path string to an existing file."""
    validate((arg, str, 0, 255))
    if not os.path.isfile(arg):
        raise ValueError("{0} does not exist.".format(arg))
    return arg


def validate_existing_directory(arg):
    """Validates an input argument is a path string to an existing directory."""
    arg = os.path.abspath(arg)
    validate((arg, str, 0, 255))
    if not os.path.isdir(arg):
        raise ValueError("{0} does not exist".format(arg))
    return arg


def validate_existing_path(arg):
    """Validates an input argument is a path string to an existing file or directory."""
    arg = os.path.abspath(arg)
    validate((arg, str, 0, 255))
    if not os.path.exists(arg):
        raise ValueError("{0} does not exist".format(arg))
    return arg


def validate_parent_exists(arg):
    """Validates an input argument is a path string, and its parent directory exists."""
    arg = os.path.abspath(arg)
    dir_arg = os.path.dirname(os.path.abspath(arg))
    if validate_existing_directory(dir_arg):
        return arg
    return None


def valid_path_append(path, *args):
    """
    Helper to validate passed path directory and append any subsequent
    filename arguments.

    Arguments:
        path (str): Initial filesystem path.  Should expand to a valid
                    directory.
        *args (list, optional): Any filename or path suffices to append to path
                                for returning.
        Returns:
            (list, str): path prepended list of files from args, or path alone if
                     no args specified.
    Raises:
        ValueError: if path is not a valid directory on this filesystem.
    """
    full_path = os.path.expanduser(path)
    res = []
    if not os.path.exists(full_path):
        os.makedirs(full_path)
    if not os.path.isdir(full_path):
        raise ValueError("path: {0} is not a valid directory".format(path))
    for suffix_path in args:
        res.append(os.path.join(full_path, suffix_path))
    if len(res) == 0:
        return path
    if len(res) == 1:
        return res[0]
    return res


def sanitize_path(path):
    s_path = os.path.normpath('/' + path).lstrip('/')
    assert len(s_path) < 255
    return s_path


def check(validator):
    class CustomAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            validator(values)
            setattr(namespace, self.dest, values)

    return CustomAction


def check_size(min_size=None, max_size=None):
    class CustomAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            validate((values, self.type, min_size, max_size, self.dest))
            setattr(namespace, self.dest, values)

    return CustomAction


def validate_proxy_path(arg):
    """Validates an input argument is a valid proxy path or None"""
    proxy_validation_regex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
    if arg is not None and re.match(proxy_validation_regex, arg) is None:
        raise ValueError("{0} is not a valid proxy path".format(arg))
    return arg


def validate_boolean(arg):
    """Validates an input argument of type boolean"""
    if arg.lower() not in ['true', 'false']:
        raise argparse.ArgumentTypeError('expected true | false argument')
    return arg.lower() == "true"


def load_json_file(file_path):
    """load a file into a json object"""
    try:
        with open(file_path) as small_file:
            return json.load(small_file)
    except OSError as e:
        print(e)
        print('trying to read file in blocks')
        with open(file_path) as big_file:
            json_string = ''
            while True:
                block = big_file.read(64 * (1 << 20))  # Read 64 MB at a time;
                json_string = json_string + block
                if not block:  # Reached EOF
                    break
            return json.loads(json_string)


def json_dumper(obj):
    """for objects that have members that cant be serialized and implement toJson() method"""
    try:
        return obj.toJson()
    except Exception:
        return obj.__dict__


def load_files_from_path(dir_path, extension='txt'):
    """load all files from given directory (with given extension)"""
    files = [os.path.join(dir_path, f) for f in os.listdir(dir_path)
             if os.path.isfile(os.path.join(dir_path, f)) and f.endswith(extension)]
    files_data = []
    for f in files:
        with open(f) as fp:
            files_data.append(' '.join(map(str.strip, fp.readlines())))
    return files_data


def create_folder(path):
    if path:
        if not os.path.exists(path):
            os.makedirs(path)


def download_unzip(url: str, sourcefile: str, unzipped_path: str or PathLike,
                   license_msg: str = None):
    """Downloads a zip file, extracts it to destination, deletes the zip file. If license_msg is
    supplied, user is prompted for download confirmation."""
    dest_parent = Path(unzipped_path).parent

    if not os.path.exists(unzipped_path):
        if license_msg is None or license_prompt(license_msg, urlparse(url).netloc):
            zip_path = dest_parent / sourcefile
            makedirs(dest_parent, exist_ok=True)
            download_unlicensed_file(url, sourcefile, zip_path)
            print('Unzipping...')
            uncompress_file(zip_path, dest_parent)
            remove(zip_path)
    return unzipped_path



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import argparse
import json
import logging

from nlp_architect.data.cdc_resources.relations.relation_types_enums import WikipediaSearchMethod
from nlp_architect.data.cdc_resources.relations.wikipedia_relation_extraction import \
    WikipediaRelationExtraction
from nlp_architect.models.cross_doc_coref.system.cdc_utils import load_mentions_vocab_from_files
from nlp_architect.utils import io
from nlp_architect.utils.io import json_dumper

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

result_dump = {}

parser = argparse.ArgumentParser(description='Create Wikipedia dataset only dump')
parser.add_argument('--mentions', type=str, help='mentions_file file', required=True)
parser.add_argument('--host', type=str, help='elastic host')
parser.add_argument('--port', type=int, help='elastic port')
parser.add_argument('--index', type=str, help='elastic index')
parser.add_argument('--output', type=str, help='location were to create dump file', required=True)

args = parser.parse_args()


def wiki_dump_from_gs():
    logger.info('Starting, process will connect with ElasticSearch and online wikipedia site...')
    mentions_files = [args.mentions]
    dump_file = args.output
    vocab = load_mentions_vocab_from_files(mentions_files)

    if args.host and args.port and args.index:
        wiki_elastic = WikipediaRelationExtraction(WikipediaSearchMethod.ELASTIC,
                                                   host=args.host,
                                                   port=args.port,
                                                   index=args.index)
    else:
        logger.info(
            'Running without Wikipedia elastic search, Note that this will '
            'take much longer to process only using online service')
        wiki_elastic = None

    wiki_online = WikipediaRelationExtraction(WikipediaSearchMethod.ONLINE)

    for phrase in vocab:
        phrase = phrase.replace("'", "").replace('"', "").replace('\\', "").strip()
        logger.info('Try to retrieve \'%s\' from elastic search', phrase)
        pages = None
        if wiki_elastic:
            pages = wiki_elastic.get_phrase_related_pages(phrase)
        if not pages or not pages.get_pages() or len(pages.get_pages()) == 0:
            logger.info('Not on elastic, retrieve \'%s\' from wiki online site', phrase)
            pages = wiki_online.get_phrase_related_pages(phrase)
        for search_page in pages.get_pages():
            add_page(search_page, phrase)

    with open(dump_file, 'w') as myfile:
        json.dump(result_dump, myfile, default=json_dumper)

    logger.info('Saving dump to file-%s', dump_file)


def add_page(search_page, phrase):
    try:
        if search_page is not None:
            if phrase not in result_dump:
                result_dump[phrase] = []
                result_dump[phrase].append(search_page)
            else:
                pages = result_dump[phrase]
                for page in pages:
                    if page.pageid == search_page.pageid:
                        return

                result_dump[phrase].append(search_page)

            logger.info('page-%s added', str(search_page))
    except Exception:
        logger.error('could not extract wiki info from phrase-%s', search_page.orig_phrase)


if __name__ == '__main__':
    io.validate_existing_filepath(args.mentions)
    io.validate_existing_filepath(args.output)
    if args.host:
        io.validate((args.host, str, 1, 1000))
    if args.port:
        io.validate((args.port, int, 1, 65536))
    if args.index:
        io.validate((args.index, str, 1, 10000))

    wiki_dump_from_gs()
